# pylint: disable=R0912,R0914,C0302
"""
@file
@brief Text representations of graphs.
"""
import pprint
from collections import OrderedDict
import numpy
from onnx import TensorProto, AttributeProto
from onnx.numpy_helper import to_array
from onnx.mapping import TENSOR_TYPE_TO_NP_TYPE
from ..tools.graphs import onnx2bigraph
from ..onnx_tools.onnx2py_helper import _var_as_dict, get_tensor_shape


def onnx_text_plot(model_onnx, recursive=False, graph_type='basic',
                   grid=5, distance=5):
    """
    Uses @see fn onnx2bigraph to convert the ONNX graph
    into text.

    :param model_onnx: onnx representation
    :param recursive: @see fn onnx2bigraph
    :param graph_type: @see fn onnx2bigraph
    :param grid: @see me display_structure
    :param distance: @see fn display_structure
    :return: text

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        import numpy
        from mlprodict.onnx_conv import to_onnx
        from mlprodict import __max_supported_opset__ as opv
        from mlprodict.plotting.plotting import onnx_text_plot
        from mlprodict.npy.xop import loadop

        OnnxAdd, OnnxSub = loadop('Add', 'Sub')

        idi = numpy.identity(2).astype(numpy.float32)
        A = OnnxAdd('X', idi, op_version=opv)
        B = OnnxSub(A, 'W', output_names=['Y'], op_version=opv)
        onx = B.to_onnx({'X': idi, 'W': idi})
        print(onnx_text_plot(onx))
    """
    bigraph = onnx2bigraph(model_onnx)
    graph = bigraph.display_structure()
    return graph.to_text()


def onnx_text_plot_tree(node):
    """
    Gives a textual representation of a tree ensemble.

    :param node: `TreeEnsemble*`
    :return: text

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        import numpy
        from sklearn.datasets import load_iris
        from sklearn.tree import DecisionTreeRegressor
        from mlprodict.onnx_conv import to_onnx
        from mlprodict.plotting.plotting import onnx_text_plot_tree

        iris = load_iris()
        X, y = iris.data.astype(numpy.float32), iris.target
        clr = DecisionTreeRegressor(max_depth=3)
        clr.fit(X, y)
        onx = to_onnx(clr, X)
        res = onnx_text_plot_tree(onx.graph.node[0])
        print(res)
    """
    def rule(r):
        if r == b'BRANCH_LEQ':
            return '<='
        if r == b'BRANCH_LT':  # pragma: no cover
            return '<'
        if r == b'BRANCH_GEQ':  # pragma: no cover
            return '>='
        if r == b'BRANCH_GT':  # pragma: no cover
            return '>'
        if r == b'BRANCH_EQ':  # pragma: no cover
            return '=='
        if r == b'BRANCH_NEQ':  # pragma: no cover
            return '!='
        raise ValueError(  # pragma: no cover
            f"Unexpected rule {rule!r}.")

    class Node:
        "Node representation."

        def __init__(self, i, atts):
            self.nodes_hitrates = None
            self.nodes_missing_value_tracks_true = None
            for k, v in atts.items():
                if k.startswith('nodes'):
                    setattr(self, k, v[i])
            self.depth = 0
            self.true_false = ''

        def process_node(self):
            "node to string"
            if self.nodes_modes == b'LEAF':  # pylint: disable=E1101
                text = "%s y=%r f=%r i=%r" % (
                    self.true_false,
                    self.target_weights, self.target_ids,  # pylint: disable=E1101
                    self.target_nodeids)  # pylint: disable=E1101
            else:
                text = "%s X%d %s %r" % (
                    self.true_false,
                    self.nodes_featureids,  # pylint: disable=E1101
                    rule(self.nodes_modes),  # pylint: disable=E1101
                    self.nodes_values)  # pylint: disable=E1101
                if self.nodes_hitrates and self.nodes_hitrates != 1:
                    text += f" hi={self.nodes_hitrates!r}"
                if self.nodes_missing_value_tracks_true:
                    text += f" miss={self.nodes_missing_value_tracks_true!r}"
            return f"{'   ' * self.depth}{text}"

    def process_tree(atts, treeid):
        "tree to string"
        rows = [f'treeid={treeid!r}']
        if 'base_values' in atts:
            if treeid < len(atts['base_values']):
                rows.append(f"base_value={atts['base_values'][treeid]!r}")

        short = {}
        for prefix in ['nodes', 'target', 'class']:
            if (f'{prefix}_treeids') not in atts:
                continue
            idx = [i for i in range(len(atts[f'{prefix}_treeids']))
                   if atts[f'{prefix}_treeids'][i] == treeid]
            for k, v in atts.items():
                if k.startswith(prefix):
                    if 'classlabels' in k:
                        short[k] = list(v)
                    else:
                        short[k] = [v[i] for i in idx]

        nodes = OrderedDict()
        for i in range(len(short['nodes_treeids'])):
            nodes[i] = Node(i, short)
        prefix = 'target' if 'target_treeids' in short else 'class'
        for i in range(len(short[f'{prefix}_treeids'])):
            idn = short[f'{prefix}_nodeids'][i]
            node = nodes[idn]
            node.target_nodeids = idn
            node.target_ids = short[f'{prefix}_ids'][i]
            node.target_weights = short[f'{prefix}_weights'][i]

        def iterate(nodes, node, depth=0, true_false=''):
            node.depth = depth
            node.true_false = true_false
            yield node
            if node.nodes_falsenodeids > 0:
                for n in iterate(nodes, nodes[node.nodes_falsenodeids],
                                 depth=depth + 1, true_false='F'):
                    yield n
                for n in iterate(nodes, nodes[node.nodes_truenodeids],
                                 depth=depth + 1, true_false='T'):
                    yield n

        for node in iterate(nodes, nodes[0]):
            rows.append(node.process_node())
        return rows

    if node.op_type in ("TreeEnsembleRegressor", "TreeEnsembleClassifier"):
        d = {k: v['value'] for k, v in _var_as_dict(node)['atts'].items()}
        atts = {}
        for k, v in d.items():
            atts[k] = v if isinstance(v, int) else list(v)
        trees = list(sorted(set(atts['nodes_treeids'])))
        if 'n_targets' in atts:
            rows = [f"n_targets={atts['n_targets']!r}"]
        else:
            rows = ['n_classes=%r' % len(
                atts.get('classlabels_int64s',
                         atts.get('classlabels_strings', [])))]
        rows.append(f'n_trees={len(trees)!r}')
        for tree in trees:
            r = process_tree(atts, tree)
            rows.append('----')
            rows.extend(r)
        return "\n".join(rows)

    raise NotImplementedError(  # pragma: no cover
        f"Type {node.op_type!r} cannot be displayed.")


def _append_succ_pred(subgraphs, successors, predecessors, node_map, node, prefix="",
                      parent_node_name=None):
    node_name = prefix + node.name + "#" + "|".join(node.output)
    node_map[node_name] = node
    successors[node_name] = []
    predecessors[node_name] = []
    for name in node.input:
        predecessors[node_name].append(name)
        if name not in successors:
            successors[name] = []
        successors[name].append(node_name)
    for name in node.output:
        successors[node_name].append(name)
        predecessors[name] = [node_name]
    if node.op_type in {'If', 'Scan', 'Loop', 'Expression'}:
        for att in node.attribute:
            if (att.type != AttributeProto.GRAPH or  # pylint: disable=E1101
                    not hasattr(att, 'g') or att.g is None):
                continue
            subgraphs.append((node, att.name, att.g))
            _append_succ_pred_s(subgraphs, successors, predecessors, node_map,
                                att.g.node, prefix=node_name + "::",
                                parent_node_name=node_name,
                                parent_graph=att.g)


def _append_succ_pred_s(subgraphs, successors, predecessors, node_map, nodes, prefix="",
                        parent_node_name=None, parent_graph=None):
    for node in nodes:
        _append_succ_pred(subgraphs, successors, predecessors, node_map, node,
                          prefix=prefix, parent_node_name=parent_node_name)
    if parent_node_name is not None:
        unknown = set()
        known = {}
        for i in parent_graph.initializer:
            known[i.name] = None
        for i in parent_graph.input:
            known[i.name] = None
        for n in parent_graph.node:
            for i in n.input:
                if i not in known:
                    unknown.add(i)
            for i in n.output:
                known[i] = n
        if len(unknown) > 0:
            # These inputs are coming from the graph below.
            for name in unknown:
                successors[name].append(parent_node_name)
                predecessors[parent_node_name].append(name)


def graph_predecessors_and_successors(graph):
    """
    Returns the successors and the predecessors within on ONNX graph.
    """
    node_map = {}
    successors = {}
    predecessors = {}
    subgraphs = []
    _append_succ_pred_s(subgraphs, successors,
                        predecessors, node_map, graph.node)
    return subgraphs, predecessors, successors, node_map


def get_hidden_inputs(nodes):
    """
    Returns the list of hidden inputs used by subgraphs.

    :param nodes: list of nodes
    :return: list of names
    """
    inputs = set()
    outputs = set()
    for node in nodes:
        inputs |= set(node.input)
        outputs |= set(node.output)
        for att in node.attribute:
            if (att.type != AttributeProto.GRAPH or  # pylint: disable=E1101
                    not hasattr(att, 'g') or att.g is None):
                continue
            hidden = get_hidden_inputs(att.g.node)
            inits = set(att.g.initializer)
            inputs |= hidden - (inits & hidden)
    return inputs - (outputs & inputs)


def reorder_nodes_for_display(nodes, verbose=False):
    """
    Reorders the node with breadth first seach (BFS).

    :param nodes: list of ONNX nodes
    :param verbose: dislay intermediate informations
    :return: reordered list of nodes
    """
    class temp:
        "Fake GraphProto."

        def __init__(self, nodes):
            self.node = nodes

    _, predecessors, successors, dnodes = graph_predecessors_and_successors(
        temp(nodes))
    local_variables = get_hidden_inputs(nodes)

    all_outputs = set()
    all_inputs = set(local_variables)
    for node in nodes:
        all_outputs |= set(node.output)
        all_inputs |= set(node.input)
    common = all_outputs & all_inputs

    successors = {k: set(v) for k, v in successors.items()}
    predecessors = {k: set(v) for k, v in predecessors.items()}
    if verbose:
        pprint.pprint(  # pragma: no cover
            ["[reorder_nodes_for_display]", "predecessors",
             predecessors, "successors", successors])

    known = all_inputs - common
    new_nodes = []
    done = set()

    def _find_sequence(node_name, known, done):
        inputs = dnodes[node_name].input
        if any(map(lambda i: i not in known, inputs)):
            return []

        res = [node_name]
        while res[-1] in successors:
            next_names = successors[res[-1]]
            if res[-1] not in dnodes:
                next_names = set(v for v in next_names if v not in known)
                if len(next_names) == 1:
                    next_name = next_names.pop()
                    inputs = dnodes[next_name].input
                    if any(map(lambda i: i not in known, inputs)):
                        break
                    res.extend(next_name)
                else:
                    break
            else:
                next_names = set(v for v in next_names if v not in done)
                if len(next_names) == 1:
                    next_name = next_names.pop()
                    res.append(next_name)
                else:
                    break

        return [r for r in res if r in dnodes and r not in done]

    while len(done) < len(nodes):
        # possible
        possibles = OrderedDict()
        for k, v in dnodes.items():
            if k in done:
                continue
            if '::' in k:
                # node part of a sub graph (assuming :: is never used in a node name)
                continue
            if predecessors[k] <= known:
                possibles[k] = v

        sequences = OrderedDict()
        for k, v in possibles.items():
            if k in done:
                continue
            sequences[k] = _find_sequence(k, known, done)
            if verbose:
                print("[reorder_nodes_for_display] * sequence(%s)=%s - %r" % (
                    k, ",".join(sequences[k]), list(sequences)))

        if len(sequences) == 0:
            raise RuntimeError(  # pragma: no cover
                "Unexpected empty sequence (len(possibles)=%d, "
                "len(done)=%d, len(nodes)=%d). This is usually due to "
                "a name used both as result name and node node. "
                "known=%r." % (len(possibles), len(done), len(nodes), known))

        # find the best sequence
        best = None
        for k, v in sequences.items():
            if best is None or len(v) > len(sequences[best]):
                # if the sequence of successors is longer
                best = k
            elif len(v) == len(sequences[best]):
                if len(new_nodes) > 0:
                    # then choose the next successor sharing input with
                    # previous output
                    so = set(new_nodes[-1].output)
                    first1 = dnodes[sequences[best][0]]
                    first2 = dnodes[v[0]]
                    if len(set(first1.input) & so) < len(set(first2.input) & so):
                        best = k
                else:
                    first1 = dnodes[sequences[best][0]]
                    first2 = dnodes[v[0]]
                    if first1.op_type > first2.op_type:
                        best = k
                    elif (first1.op_type == first2.op_type and
                            first1.name > first2.name):
                        best = k

        if best is None:
            raise RuntimeError(  # pragma: no cover
                f"Wrong implementation (len(sequence)={len(sequences)}).")
        if verbose:
            print("[reorder_nodes_for_display] BEST: sequence(%s)=%s" % (
                best, ",".join(sequences[best])))

        # process the sequence
        for k in sequences[best]:
            v = dnodes[k]
            new_nodes.append(v)
            if verbose:
                print(
                    f"[reorder_nodes_for_display] + {v.name!r} ({v.op_type!r})")
            done.add(k)
            known |= set(v.output)

    if len(new_nodes) != len(nodes):
        raise RuntimeError(  # pragma: no cover
            "The returned new nodes are different. "
            "len(nodes=%d) != %d=len(new_nodes). done=\n%r"
            "\n%s\n----------\n%s" % (
                len(nodes), len(new_nodes), done,
                "\n".join("%d - %s - %s - %s" % (
                    (n.name + "".join(n.output)) in done,
                    n.op_type, n.name, n.name + "".join(n.output))
                    for n in nodes),
                "\n".join("%d - %s - %s - %s" % (
                    (n.name + "".join(n.output)) in done,
                    n.op_type, n.name, n.name + "".join(n.output))
                    for n in new_nodes)))
    n0s = set(n.name for n in nodes)
    n1s = set(n.name for n in new_nodes)
    if n0s != n1s:
        raise RuntimeError(  # pragma: no cover
            "The returned new nodes are different.\n"
            "%r !=\n%r\ndone=\n%r"
            "\n----------\n%s\n----------\n%s" % (
                n0s, n1s, done,
                "\n".join("%d - %s - %s - %s" % (
                    (n.name + "".join(n.output)) in done,
                    n.op_type, n.name, n.name + "".join(n.output))
                    for n in nodes),
                "\n".join("%d - %s - %s - %s" % (
                    (n.name + "".join(n.output)) in done,
                    n.op_type, n.name, n.name + "".join(n.output))
                    for n in new_nodes)))
    return new_nodes


def _get_type(obj0):
    obj = obj0
    if hasattr(obj, 'data_type'):
        if (obj.data_type == TensorProto.FLOAT and  # pylint: disable=E1101
                hasattr(obj, 'float_data')):
            return TENSOR_TYPE_TO_NP_TYPE[TensorProto.FLOAT]  # pylint: disable=E1101
        if (obj.data_type == TensorProto.DOUBLE and  # pylint: disable=E1101
                hasattr(obj, 'double_data')):
            return TENSOR_TYPE_TO_NP_TYPE[TensorProto.DOUBLE]  # pylint: disable=E1101
        if (obj.data_type == TensorProto.INT64 and  # pylint: disable=E1101
                hasattr(obj, 'int64_data')):
            return TENSOR_TYPE_TO_NP_TYPE[TensorProto.INT64]  # pylint: disable=E1101
        if (obj.data_type == TensorProto.INT32 and  # pylint: disable=E1101
                hasattr(obj, 'int32_data')):
            return TENSOR_TYPE_TO_NP_TYPE[TensorProto.INT32]  # pylint: disable=E1101
        raise RuntimeError(  # pragma: no cover
            f"Unable to guess type from {obj0!r}.")
    if hasattr(obj, 'type'):
        obj = obj.type
    if hasattr(obj, 'tensor_type'):
        obj = obj.tensor_type
    if hasattr(obj, 'elem_type'):
        return TENSOR_TYPE_TO_NP_TYPE.get(obj.elem_type, '?')
    raise RuntimeError(  # pragma: no cover
        f"Unable to guess type from {obj0!r}.")


def _get_shape(obj):
    obj0 = obj
    if hasattr(obj, 'data_type'):
        if (obj.data_type == TensorProto.FLOAT and  # pylint: disable=E1101
                hasattr(obj, 'float_data')):
            return (len(obj.float_data), )
        if (obj.data_type == TensorProto.DOUBLE and  # pylint: disable=E1101
                hasattr(obj, 'double_data')):
            return (len(obj.double_data), )
        if (obj.data_type == TensorProto.INT64 and  # pylint: disable=E1101
                hasattr(obj, 'int64_data')):
            return (len(obj.int64_data), )
        if (obj.data_type == TensorProto.INT32 and  # pylint: disable=E1101
                hasattr(obj, 'int32_data')):
            return (len(obj.int32_data), )
        raise RuntimeError(  # pragma: no cover
            f"Unable to guess type from {obj0!r}.")
    if hasattr(obj, 'type'):
        obj = obj.type
    if hasattr(obj, 'tensor_type'):
        return get_tensor_shape(obj)
    raise RuntimeError(  # pragma: no cover
        f"Unable to guess type from {obj0!r}.")


def onnx_simple_text_plot(model, verbose=False, att_display=None,  # pylint: disable=R0915
                          add_links=False, recursive=False, functions=True,
                          raise_exc=True, sub_graphs_names=None,
                          level=1, indent=True):
    """
    Displays an ONNX graph into text.

    :param model: ONNX graph
    :param verbose: display debugging information
    :param att_display: list of attributes to display, if None,
        a default list if used
    :param add_links: displays links of the right side
    :param recursive: display subgraphs as well
    :param functions: display functions as well
    :param raise_exc: raises an exception if the model is not valid,
        otherwise tries to continue
    :param sub_graphs_names: list of sub-graphs names
    :param level: sub-graph level
    :param indent: use indentation or not
    :return: str

    An ONNX graph is printed the following way:

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        import numpy
        from sklearn.cluster import KMeans
        from mlprodict.plotting.plotting import onnx_simple_text_plot
        from mlprodict.onnx_conv import to_onnx

        x = numpy.random.randn(10, 3)
        y = numpy.random.randn(10)
        model = KMeans(3)
        model.fit(x, y)
        onx = to_onnx(model, x.astype(numpy.float32),
                      target_opset=15)
        text = onnx_simple_text_plot(onx, verbose=False)
        print(text)

    The same graphs with links.

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        import numpy
        from sklearn.cluster import KMeans
        from mlprodict.plotting.plotting import onnx_simple_text_plot
        from mlprodict.onnx_conv import to_onnx

        x = numpy.random.randn(10, 3)
        y = numpy.random.randn(10)
        model = KMeans(3)
        model.fit(x, y)
        onx = to_onnx(model, x.astype(numpy.float32),
                      target_opset=15)
        text = onnx_simple_text_plot(onx, verbose=False, add_links=True)
        print(text)

    Visually, it looks like the following:

    .. gdot::
        :script: DOT-SECTION

        import numpy
        from sklearn.cluster import KMeans
        from mlprodict.onnxrt import OnnxInference
        from mlprodict.onnx_conv import to_onnx

        x = numpy.random.randn(10, 3)
        y = numpy.random.randn(10)
        model = KMeans(3)
        model.fit(x, y)
        model_onnx = to_onnx(model, x.astype(numpy.float32),
                             target_opset=15)
        oinf = OnnxInference(model_onnx, inplace=False)

        print("DOT-SECTION", oinf.to_dot())
    """
    use_indentation = indent
    if att_display is None:
        att_display = [
            'activations',
            'align_corners',
            'allowzero',
            'alpha',
            'auto_pad',
            'axis',
            'axes',
            'batch_axis',
            'batch_dims',
            'beta',
            'bias',
            'blocksize',
            'case_change_action',
            'ceil_mode',
            'center_point_box',
            'clip',
            'coordinate_transformation_mode',
            'count_include_pad',
            'cubic_coeff_a',
            'decay_factor',
            'detect_negative',
            'detect_positive',
            'dilation',
            'dilations',
            'direction',
            'dtype',
            'end',
            'epsilon',
            'equation',
            'exclusive',
            'exclude_outside',
            'extrapolation_value',
            'fmod',
            'gamma',
            'group',
            'hidden_size',
            'high',
            'ignore_index',
            'input_forget',
            'is_case_sensitive',
            'k',
            'keepdims',
            'kernel_shape',
            'lambd',
            'largest',
            'layout',
            'linear_before_reset',
            'locale',
            'low',
            'max_gram_length',
            'max_skip_count',
            'mean',
            'min_gram_length',
            'mode',
            'momentum',
            'nearest_mode',
            'ngram_counts',
            'ngram_indexes',
            'noop_with_empty_axes',
            'norm_coefficient',
            'norm_coefficient_post',
            'num_scan_inputs',
            'output_height',
            'output_padding',
            'output_shape',
            'output_width',
            'p',
            'padding_mode',
            'pads',
            'perm',
            'pooled_shape',
            'reduction',
            'reverse',
            'sample_size',
            'sampling_ratio',
            'scale',
            'scan_input_axes',
            'scan_input_directions',
            'scan_output_axes',
            'scan_output_directions',
            'seed',
            'select_last_index',
            'size',
            'sorted',
            'spatial_scale',
            'start',
            'storage_order',
            'strides',
            'time_axis',
            'to',
            'training_mode',
            'transA',
            'transB',
            'type',
            'upper',
            'xs',
            'y',
            'zs',
        ]

    if sub_graphs_names is None:
        sub_graphs_names = {}

    def _get_subgraph_name(idg):
        if idg in sub_graphs_names:
            return sub_graphs_names[idg]
        g = "G%d" % (len(sub_graphs_names) + 1)
        sub_graphs_names[idg] = g
        return g

    def str_node(indent, node):
        atts = []
        if hasattr(node, 'attribute'):
            for att in node.attribute:
                done = True
                if att.name in att_display:
                    if att.type == AttributeProto.INT:  # pylint: disable=E1101
                        atts.append("%s=%d" % (att.name, att.i))
                    elif att.type == AttributeProto.FLOAT:  # pylint: disable=E1101
                        atts.append(f"{att.name}={att.f:1.2f}")
                    elif att.type == AttributeProto.INTS:  # pylint: disable=E1101
                        atts.append("%s=%s" % (att.name, str(
                            list(att.ints)).replace(" ", "")))
                    else:
                        done = False
                elif (att.type == AttributeProto.GRAPH and  # pylint: disable=E1101
                        hasattr(att, 'g') and att.g is not None):
                    atts.append(f"{att.name}={_get_subgraph_name(id(att.g))}")
                elif att.ref_attr_name:
                    atts.append(f"{att.name}=${att.ref_attr_name}")
                else:
                    done = False
                if done:
                    continue
                if att.type in (AttributeProto.TENSOR,  # pylint: disable=E1101
                                AttributeProto.TENSORS,  # pylint: disable=E1101
                                AttributeProto.SPARSE_TENSOR,  # pylint: disable=E1101
                                AttributeProto.SPARSE_TENSORS):  # pylint: disable=E1101
                    try:
                        val = str(to_array(att.t).tolist())
                    except TypeError as e:  # pragma: no cover
                        raise TypeError(
                            "Unable to display tensor type %r.\n%s" % (
                                att.type, str(att))) from e
                    if "\n" in val:
                        val = val.split("\n", maxsplit=1) + "..."
                    if len(val) > 10:
                        val = val[:10] + "..."
                elif att.type == AttributeProto.STRING:  # pylint: disable=E1101
                    val = str(att.s)
                elif att.type == AttributeProto.STRINGS:  # pylint: disable=E1101
                    n_val = list(att.strings)
                    if len(n_val) < 5:
                        val = ",".join(map(str, n_val))
                    else:
                        val = "%d:[%s...%s]" % (
                            len(n_val),
                            ",".join(map(str, n_val[:2])),
                            ",".join(map(str, n_val[-2:])))
                elif att.type == AttributeProto.INT:  # pylint: disable=E1101
                    val = str(att.i)
                elif att.type == AttributeProto.FLOAT:  # pylint: disable=E1101
                    val = str(att.f)
                elif att.type == AttributeProto.INTS:  # pylint: disable=E1101
                    n_val = list(att.ints)
                    if len(n_val) < 6:
                        val = f"[{','.join(map(str, n_val))}]"
                    else:
                        val = "%d:[%s...%s]" % (
                            len(n_val),
                            ",".join(map(str, n_val[:3])),
                            ",".join(map(str, n_val[-3:])))
                elif att.type == AttributeProto.FLOATS:  # pylint: disable=E1101
                    n_val = list(att.floats)
                    if len(n_val) < 5:
                        val = f"[{','.join(map(str, n_val))}]"
                    else:
                        val = "%d:[%s...%s]" % (
                            len(n_val),
                            ",".join(map(str, n_val[:2])),
                            ",".join(map(str, n_val[-2:])))
                else:
                    val = '.%d' % att.type
                atts.append(f"{att.name}={val}")
        inputs = list(node.input)
        if len(atts) > 0:
            inputs.extend(atts)
        if node.domain in ('', 'ai.onnx.ml'):
            domain = ''
        else:
            domain = f'[{node.domain}]'
        return "%s%s%s(%s) -> %s" % (
            "  " * indent, node.op_type, domain,
            ", ".join(inputs), ", ".join(node.output))

    rows = []
    if hasattr(model, 'opset_import'):
        for opset in model.opset_import:
            rows.append(
                f"opset: domain={opset.domain!r} version={opset.version!r}")
    if hasattr(model, 'graph'):
        if model.doc_string:
            rows.append(f'doc_string: {model.doc_string}')
        main_model = model
        model = model.graph
    else:
        main_model = None

    # inputs
    line_name_new = {}
    line_name_in = {}
    if level == 0:
        rows.append("----- input ----")
    for inp in model.input:
        if isinstance(inp, str):
            rows.append(f"input: {inp!r}")
        else:
            line_name_new[inp.name] = len(rows)
            rows.append("input: name=%r type=%r shape=%r" % (
                inp.name, _get_type(inp), _get_shape(inp)))
    if hasattr(model, 'attribute'):
        for att in model.attribute:
            if isinstance(att, str):
                rows.append(f"attribute: {att!r}")
            else:
                raise NotImplementedError(  # pragma: no cover
                    "Not yet introduced in onnx.")

    # initializer
    if hasattr(model, 'initializer'):
        if len(model.initializer) and level == 0:
            rows.append("----- initializer ----")
        for init in model.initializer:
            if numpy.prod(_get_shape(init)) < 5:
                content = f" -- {to_array(init).ravel()!r}"
            else:
                content = ""
            line_name_new[init.name] = len(rows)
            rows.append("init: name=%r type=%r shape=%r%s" % (
                init.name, _get_type(init), _get_shape(init), content))
    if level == 0:
        rows.append("----- main graph ----")

    # successors, predecessors, it needs to support subgraphs
    subgraphs = graph_predecessors_and_successors(model)[0]

    # walk through nodes
    init_names = set()
    indents = {}
    for inp in model.input:
        if isinstance(inp, str):
            indents[inp] = 0
            init_names.add(inp)
        else:
            indents[inp.name] = 0
            init_names.add(inp.name)
    if hasattr(model, 'initializer'):
        for init in model.initializer:
            indents[init.name] = 0
            init_names.add(init.name)

    try:
        nodes = reorder_nodes_for_display(model.node, verbose=verbose)
    except RuntimeError as e:  # pragma: no cover
        if raise_exc:
            raise e
        else:
            rows.append(f"ERROR: {e}")
        nodes = model.node

    previous_indent = None
    previous_out = None
    previous_in = None
    for node in nodes:
        add_break = False
        name = node.name + "#" + "|".join(node.output)
        if name in indents:
            indent = indents[name]
            if previous_indent is not None and indent < previous_indent:
                if verbose:
                    print(f"[onnx_simple_text_plot] break1 {node.op_type}")
                add_break = True
        elif previous_in is not None and set(node.input) == previous_in:
            indent = previous_indent
        else:
            inds = [indents.get(i, 0)
                    for i in node.input if i not in init_names]
            if len(inds) == 0:
                indent = 0
            else:
                mi = min(inds)
                indent = mi
                if previous_indent is not None and indent < previous_indent:
                    if verbose:
                        print(  # pragma: no cover
                            f"[onnx_simple_text_plot] break2 {node.op_type}")
                    add_break = True
            if not add_break and previous_out is not None:
                if len(set(node.input) & previous_out) == 0:
                    if verbose:
                        print(f"[onnx_simple_text_plot] break3 {node.op_type}")
                    add_break = True
                    indent = 0

        if add_break and verbose:
            print("[onnx_simple_text_plot] add break")
        for n in node.input:
            if n in line_name_in:
                line_name_in[n].append(len(rows))
            else:
                line_name_in[n] = [len(rows)]
        for n in node.output:
            line_name_new[n] = len(rows)
        rows.append(str_node(indent if use_indentation else 0, node))
        indents[name] = indent

        for i, o in enumerate(node.output):
            indents[o] = indent + 1

        previous_indent = indents[name]
        previous_out = set(node.output)
        previous_in = set(node.input)

    # outputs
    if level == 0:
        rows.append("----- output ----")
    for out in model.output:
        if isinstance(out, str):
            if out in line_name_in:
                line_name_in[out].append(len(rows))
            else:
                line_name_in[out] = [len(rows)]
            rows.append(f"output: name={out!r} type={'?'} shape={'?'}")
        else:
            if out.name in line_name_in:
                line_name_in[out.name].append(len(rows))
            else:
                line_name_in[out.name] = [len(rows)]
            rows.append("output: name=%r type=%r shape=%r" % (
                out.name, _get_type(out), _get_shape(out)))

    if add_links:

        def _mark_link(rows, lengths, r1, r2, d):
            maxl = max(lengths[r1], lengths[r2]) + d * 2
            maxl = max(maxl, max(len(rows[r]) for r in range(r1, r2 + 1))) + 2

            if rows[r1][-1] == '|':
                p1, p2 = rows[r1][:lengths[r1] + 2], rows[r1][lengths[r1] + 2:]
                rows[r1] = p1 + p2.replace(' ', '-')
            rows[r1] += ("-" * (maxl - len(rows[r1]) - 1)) + "+"

            if rows[r2][-1] == " ":
                rows[r2] += "<"
            elif rows[r2][-1] == '|':
                if "<" not in rows[r2]:
                    p = lengths[r2]
                    rows[r2] = rows[r2][:p] + '<' + rows[r2][p + 1:]
                p1, p2 = rows[r2][:lengths[r2] + 2], rows[r2][lengths[r2] + 2:]
                rows[r2] = p1 + p2.replace(' ', '-')
            rows[r2] += ("-" * (maxl - len(rows[r2]) - 1)) + "+"

            for r in range(r1 + 1, r2):
                if len(rows[r]) < maxl:
                    rows[r] += " " * (maxl - len(rows[r]) - 1)
                rows[r] += "|"

        diffs = []
        for n, r1 in line_name_new.items():
            if n not in line_name_in:
                continue
            r2s = line_name_in[n]
            for r2 in r2s:
                if r1 >= r2:
                    continue
                diffs.append((r2 - r1, (n, r1, r2)))
        diffs.sort()
        for i in range(len(rows)):  # pylint: disable=C0200
            rows[i] += "  "
        lengths = [len(r) for r in rows]

        for d, (n, r1, r2) in diffs:
            if d == 1 and len(line_name_in[n]) == 1:
                # no line for link to the next node
                continue
            _mark_link(rows, lengths, r1, r2, d)

    # subgraphs
    if recursive:
        for node, name, g in subgraphs:
            rows.append('----- subgraph ---- %s - %s - att.%s=%s -- level=%d' % (
                node.op_type, node.name, name, _get_subgraph_name(id(g)),
                level))
            res = onnx_simple_text_plot(
                g, verbose=verbose, att_display=att_display,
                add_links=add_links, recursive=recursive,
                sub_graphs_names=sub_graphs_names, level=level + 1,
                raise_exc=raise_exc)
            rows.append(res)

    # functions
    if functions and main_model is not None:
        for fct in main_model.functions:
            rows.append(f'----- function name={fct.name} domain={fct.domain}')
            if fct.doc_string:
                rows.append(f'----- doc_string: {fct.doc_string}')
            res = onnx_simple_text_plot(
                fct, verbose=verbose, att_display=att_display,
                add_links=add_links, recursive=recursive,
                functions=False, sub_graphs_names=sub_graphs_names,
                level=1)
            rows.append(res)

    return "\n".join(rows)


def onnx_text_plot_io(model, verbose=False, att_display=None):
    """
    Displays information about input and output types.

    :param model: ONNX graph
    :param verbose: display debugging information
    :return: str

    An ONNX graph is printed the following way:

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        import numpy
        from sklearn.cluster import KMeans
        from mlprodict.plotting.plotting import onnx_text_plot_io
        from mlprodict.onnx_conv import to_onnx

        x = numpy.random.randn(10, 3)
        y = numpy.random.randn(10)
        model = KMeans(3)
        model.fit(x, y)
        onx = to_onnx(model, x.astype(numpy.float32),
                      target_opset=15)
        text = onnx_text_plot_io(onx, verbose=False)
        print(text)
    """
    rows = []
    if hasattr(model, 'opset_import'):
        for opset in model.opset_import:
            rows.append(
                f"opset: domain={opset.domain!r} version={opset.version!r}")
    if hasattr(model, 'graph'):
        model = model.graph

    # inputs
    for inp in model.input:
        rows.append("input: name=%r type=%r shape=%r" % (
            inp.name, _get_type(inp), _get_shape(inp)))
    # initializer
    for init in model.initializer:
        rows.append("init: name=%r type=%r shape=%r" % (
            init.name, _get_type(init), _get_shape(init)))
    # outputs
    for out in model.output:
        rows.append("output: name=%r type=%r shape=%r" % (
            out.name, _get_type(out), _get_shape(out)))
    return "\n".join(rows)



"""
@file
@brief Computes shape inference for element wise operators.
"""
import numpy
from .shape_excs import ShapeInferenceException
from .shape_result import ShapeResult, OnnxKind


def _element_wise(known_shapes, node, return_bool=False, same_type=True,
                  one_input=False):
    """
    Infers shape for an element wise operator.
    The function returns but updates *known_shapes*.

    :param known_shapes: known shapes
    :param node: Onnx node
    :param return_bool: return boolean
    :param same_type: check the type are the same
    :param one_input: allow one input
    :return: updated or not
    """
    if one_input:
        if len(node.input) == 1:
            x = known_shapes[node.input[0]]
            return known_shapes.update(node.output[0], x.copy())
    elif len(node.input) != 2:
        raise ShapeInferenceException(  # pragma: no cover
            f"Node {node.name!r} must have two inputs not {len(node.input)}.")
    x = known_shapes[node.input[0]]
    y = known_shapes[node.input[1]]
    if x.mtype != OnnxKind.Tensor:
        raise ShapeInferenceException(  # pragma: no cover
            f"Result {x!r} must be a tensor.")
    if y.mtype != OnnxKind.Tensor:
        raise ShapeInferenceException(  # pragma: no cover
            f"Result {y!r} must be a tensor.")
    if return_bool:
        return known_shapes.update(
            node.output[0],
            ShapeResult.broadcast(
                x, y, name=node.output[0], dtype=numpy.bool_,
                same_type=same_type))
    return known_shapes.update(
        node.output[0],
        ShapeResult.broadcast(
            x, y, name=node.output[0], same_type=same_type))


def shape_add(known_shapes, node):
    "Infers shape for operator Add."
    return _element_wise(known_shapes, node)


def shape_and(known_shapes, node):
    "Infers shape for operator And."
    return _element_wise(known_shapes, node)


def shape_div(known_shapes, node):
    "Infers shape for operator Div."
    return _element_wise(known_shapes, node)


def shape_equal(known_shapes, node):
    "Infers shape for operator Equal."
    return _element_wise(known_shapes, node, return_bool=True)


def shape_greater(known_shapes, node):
    "Infers shape for operator Greater."
    return _element_wise(known_shapes, node, return_bool=True)


def shape_greaterorequal(known_shapes, node):
    "Infers shape for operator GreaterOrEqual."
    return _element_wise(known_shapes, node, return_bool=True)


def shape_less(known_shapes, node):
    "Infers shape for operator Less."
    return _element_wise(known_shapes, node, return_bool=True)


def shape_lessorequal(known_shapes, node):
    "Infers shape for operator LessOrEqual."
    return _element_wise(known_shapes, node, return_bool=True)


def shape_max(known_shapes, node):
    "Infers shape for operator Max."
    return _element_wise(known_shapes, node, one_input=True)


def shape_min(known_shapes, node):
    "Infers shape for operator Min."
    return _element_wise(known_shapes, node, one_input=True)


def shape_mod(known_shapes, node):
    "Infers shape for operator Mod."
    return _element_wise(known_shapes, node)


def shape_mul(known_shapes, node):
    "Infers shape for operator Mul."
    return _element_wise(known_shapes, node)


def shape_or(known_shapes, node):
    "Infers shape for operator Or."
    return _element_wise(known_shapes, node)


def shape_pow(known_shapes, node):
    "Infers shape for operator Pow."
    return _element_wise(known_shapes, node, same_type=False)


def shape_sub(known_shapes, node):
    "Infers shape for operator Sub."
    return _element_wise(known_shapes, node)


def shape_xor(known_shapes, node):
    "Infers shape for operator Xor."
    return _element_wise(known_shapes, node)



"""
@file
@brief Experimental implementation.
"""
from collections import OrderedDict
import numpy


def custom_pad(arr, paddings, constant=0, verbose=False):
    """
    Implements function
    `pad <https://numpy.org/doc/stable/reference/
    generated/numpy.pad.html>`_ in python,
    only the constant version.

    :param arr: array
    :param paddings: paddings
    :param constant: constant
    :return: padded array
    """
    if paddings.shape[0] != len(arr.shape):
        raise ValueError(  # pragma: no cover
            f"Input shape {arr.shape} and paddings {paddings} are inconsistent.")
    if min(paddings.ravel()) < 0:
        raise NotImplementedError("Negative paddings is not implemented yet.")
    if not arr.flags['C_CONTIGUOUS']:
        arr = numpy.ascontiguousarray(arr)  # pragma: no cover

    new_shape = tuple(
        a + s for a, s in zip(arr.shape, numpy.sum(paddings, axis=1, keepdims=0)))

    cumulative_copy = [1]
    for a in reversed(new_shape):
        cumulative_copy.insert(0, a * cumulative_copy[0])
    cumulative_input = [1]
    for a in reversed(arr.shape):
        cumulative_input.insert(0, a * cumulative_input[0])

    input_arr = arr.ravel()
    if verbose:
        res = numpy.zeros(cumulative_copy[0], dtype=arr.dtype) - 1
    else:
        res = numpy.empty(cumulative_copy[0], dtype=arr.dtype)

    # preparation
    first_index = sum(
        p * c for p, c in zip(paddings[:, 0], cumulative_copy[1:]))
    dh_input = arr.shape[-1]
    dh_copy = new_shape[-1]

    # constance
    no_constant = 1 if constant == 0 else 0
    res[first_index:cumulative_copy[0]:dh_copy] = no_constant

    # padding
    for i, sh in enumerate(new_shape):
        upper_number = cumulative_copy[0] // cumulative_copy[i]
        contiguous = cumulative_copy[i + 1]
        big_index = 0
        p_left = paddings[i, 0] * contiguous
        p_right = paddings[i, 1] * contiguous
        dp = sh * contiguous - p_right
        for _ in range(upper_number):
            if p_left > 0:
                res[big_index:big_index + p_left] = constant
            if p_right > 0:
                index = big_index + dp
                res[index:index + p_right] = constant
            big_index += cumulative_copy[i]

    # copy
    index_input = 0
    index_copy = first_index
    while index_copy < cumulative_copy[0]:
        if res[index_copy] == no_constant:
            res[index_copy:index_copy + dh_input] = \
                input_arr[index_input:index_input + dh_input]
            index_input += dh_input
        index_copy += dh_copy

    # final
    return res.reshape(new_shape)


def custom_einsum(equation, x, y, verbose=False):
    """
    Experimental implementation of operator Einsum
    when it does a matrix multiplication.
    Case: ``bsnh,btnh->bnts`` with shapes
    `(1,512,12,64)` and `(1,512,12,64)`.

    :param equation: equation
    :param x: first matrix
    :param y: second matrix
    :param verbose: display internal information
    :return: result of *einsum*

    This implementation does not any transpose,
    it does a direct computation of the final result.
    It does not implementation diagonal summation (square product).
    """
    def _check_eq(eq, sh):
        if len(eq) != len(sh):
            raise ValueError(
                f"Unable to map equation {eq!r} to shape {sh!r}.")

    def _split(eq, sh):
        dx = OrderedDict((e, (v, i)) for i, (e, v) in enumerate(zip(eq, sh)))
        return dx

    def _interpret(dx, dy, eqr):
        c_uni = []
        c_trp = []
        c_sum = []
        for r in eqr:
            if r in dx:
                if r in dy:
                    if dx[r][0] != dy[r][0]:
                        raise ValueError(
                            f"Dimension mismatch for letter {r!r} dx={dx!r} dy={dy!r}.")
                    c_trp.append(r)
                else:
                    c_uni.append((r, None))
            elif r in dy:
                c_uni.append((None, r))
            else:
                raise ValueError(  # pragma: no cover
                    f"Unexpected letter {r!r} in result {eqr!r}.")
        for c in dx:
            if c not in eqr:
                if c not in dy:
                    raise ValueError(  # pragma: no cover
                        f"Unable to guess what to do with column {c!r} (left side)")
                if dx[c][0] != dy[c][0]:
                    raise ValueError(  # pragma: no cover
                        f"Dimension mismatch for letter {c!r} dx={dx!r} dy={dy!r}.")
                c_sum.append(c)
        for c in dy:
            if c not in eqr and c not in dx:
                raise ValueError(  # pragma: no cover
                    f"Unable to guess what to do with column {c!r} (right side)")
        shape = OrderedDict()
        for i, r in enumerate(eqr):
            if r in c_trp:
                shape[r] = (dx[r][0], i)
            else:
                for a, b in c_uni:
                    if a == r:
                        shape[r] = (dx[r][0], i)
                        break
                    if b == r:
                        shape[r] = (dy[r][0], i)
                        break
        if len(shape) != len(eqr):
            raise RuntimeError(  # pragma: no cover
                "Unable to compute the output shape "
                "dx=%r dy=%r eqr=%r got shape=%r." % (dx, dy, eqr, shape))
        return shape, c_trp, c_uni, c_sum

    def _inc(d):
        t = 1
        drev = list(reversed(d.items()))
        res = []
        for c, (sh, p) in drev:
            res.append((c, (t, p)))
            t *= sh
        return OrderedDict(reversed(res))

    def prod(seq):
        p = 1
        for s in seq:
            p *= s
        return p

    def get_index(cd, shape, index, col_sum):
        ind = 0
        for c, i in zip(shape, index):
            if c in cd:
                inc = cd[c][0]
                ind += inc * i
        return ind, cd[col_sum][0]

    def get_incs(cd, shape):
        incs = []
        for c in shape:
            inc = cd[c][0] if c in cd else 0
            incs.append(inc)
        return incs

    if x.dtype != y.dtype:
        raise RuntimeError("x and y must have the same dtype.")
    eqx = equation.split(',')[0]
    eqy = equation.split(',')[-1].split('->')[0]
    eqr = equation.split('->')[-1]
    _check_eq(eqx, x.shape)
    _check_eq(eqy, y.shape)
    dx = _split(eqx, x.shape)
    dy = _split(eqy, y.shape)
    shape, __, _, c_sum = _interpret(dx, dy, eqr)
    cdx = _inc(dx)
    cdy = _inc(dy)
    xrav = x.ravel()
    yrav = y.ravel()
    full_size = prod(v[0] for v in shape.values())
    zrav = numpy.empty((full_size, ), dtype=x.dtype)

    # loop
    if len(c_sum) != 1:
        raise NotImplementedError(
            f"More than one summation indices {c_sum!r} in equation {equation!r}.")
    zeros = numpy.zeros((1, ), dtype=x.dtype)
    shape_dims = [v[0] for v in shape.values()]
    index = [0 for s in shape]
    len_index = len(index)
    loop_size = dx[c_sum[0]][0]

    i_left_loop, inc_left = get_index(cdx, shape, index, c_sum[0])
    i_right_loop, inc_right = get_index(cdy, shape, index, c_sum[0])
    left_incs = get_incs(cdx, shape)
    right_incs = get_incs(cdy, shape)

    if verbose:
        def MakeString(*args):
            return "".join(map(str, args))

        print(MakeString("equation=", equation))
        print(MakeString("c_sum=", c_sum))
        print(MakeString("full_size=", full_size))
        print(MakeString("loop_size=", loop_size))
        print(MakeString("i_left_loop=", i_left_loop))
        print(MakeString("i_right_loop=", i_right_loop))
        print(MakeString("inc_left=", inc_left))
        print(MakeString("inc_right=", inc_right))
        print(MakeString("left_incs=", left_incs))
        print(MakeString("right_incs=", right_incs))
        print(MakeString("shape=", shape))
        print(MakeString("cdx=", cdx))
        print(MakeString("cdy=", cdy))

    for i in range(0, full_size):

        i_left = i_left_loop
        i_right = i_right_loop

        # summation
        add = zeros[0]
        for _ in range(loop_size):
            add += xrav[i_left] * yrav[i_right]
            i_left += inc_left
            i_right += inc_right
        zrav[i] = add

        if verbose:
            print(MakeString(
                "  -- index=", index, " ii=", i,
                " i_left_loop=", i_left_loop, " i_right_loop=", i_right_loop,
                " add=", add))

        # increment
        pos = len_index - 1
        index[pos] += 1
        i_left_loop += left_incs[pos]
        i_right_loop += right_incs[pos]
        while pos > 0 and index[pos] >= shape_dims[pos]:
            i_left_loop -= left_incs[pos] * index[pos]
            i_right_loop -= right_incs[pos] * index[pos]
            index[pos] = 0
            pos -= 1
            index[pos] += 1
            i_left_loop += left_incs[pos]
            i_right_loop += right_incs[pos]

    new_shape = tuple(v[0] for v in shape.values())
    return zrav.reshape(new_shape)



"""
@file
@brief Alternative to dot to display a graph.

.. versionadded:: 0.7
"""
import pprint
import hashlib
import numpy
import onnx


def make_hash_bytes(data, length=20):
    """
    Creates a hash of length *length*.
    """
    m = hashlib.sha256()
    m.update(data)
    res = m.hexdigest()[:length]
    return res


class AdjacencyGraphDisplay:
    """
    Structure which contains the necessary information to
    display a graph using an adjacency matrix.

    .. versionadded:: 0.7
    """

    class Action:
        "One action to do."

        def __init__(self, x, y, kind, label, orientation=None):
            self.x = x
            self.y = y
            self.kind = kind
            self.label = label
            self.orientation = orientation

        def __repr__(self):
            "usual"
            return "%s(%r, %r, %r, %r, %r)" % (
                self.__class__.__name__,
                self.x, self.y, self.kind, self.label,
                self.orientation)

    def __init__(self):
        self.actions = []

    def __iter__(self):
        "Iterates over actions."
        for act in self.actions:
            yield act

    def __str__(self):
        "usual"
        rows = [f"{self.__class__.__name__}("]
        for act in self:
            rows.append(f"    {act!r}")
        rows.append(")")
        return "\n".join(rows)

    def add(self, x, y, kind, label, orientation=None):
        """
        Adds an action to display the graph.

        :param x: x coordinate
        :param y: y coordinate
        :param kind: `'cross'` or `'text'`
        :param label: specific to kind
        :param orientation: a 2-uple `(i,j)` where *i* or *j* in `{-1,0,1}`
        """
        if kind not in {'cross', 'text'}:
            raise ValueError(  # pragma: no cover
                f"Unexpected value for kind {kind!r}.")
        if kind == 'cross' and label[0] not in {'I', 'O'}:
            raise ValueError(  # pragma: no cover
                "kind=='cross' and label[0]=%r not in {'I','O'}." % label)
        if not isinstance(label, str):
            raise TypeError(  # pragma: no cover
                f"Unexpected label type {type(label)!r}.")
        self.actions.append(
            AdjacencyGraphDisplay.Action(x, y, kind, label=label,
                                         orientation=orientation))

    def to_text(self):
        """
        Displays the graph as a single string.
        See @see fn onnx2bigraph to see how the result
        looks like.

        :return: str
        """
        mat = {}
        for act in self:
            if act.kind == 'cross':
                if act.orientation != (1, 0):
                    raise NotImplementedError(  # pragma: no cover
                        "Orientation for 'cross' must be (1, 0) not %r."
                        "" % act.orientation)
                if len(act.label) == 1:
                    mat[act.x * 3, act.y] = act.label
                elif len(act.label) == 2:
                    mat[act.x * 3, act.y] = act.label[0]
                    mat[act.x * 3 + 1, act.y] = act.label[1]
                else:
                    raise NotImplementedError(
                        f"Unable to display long cross label ({act.label!r}).")
            elif act.kind == 'text':
                x = act.x * 3
                y = act.y
                orient = act.orientation
                charset = list(act.label if max(orient) == 1
                               else reversed(act.label))
                for c in charset:
                    mat[x, y] = c
                    x += orient[0]
                    y += orient[1]
            else:
                raise ValueError(  # pragma: no cover
                    f"Unexpected kind value {act.kind!r}.")

        min_i = min(k[0] for k in mat)
        min_j = min(k[1] for k in mat)
        mat2 = {}
        for k, v in mat.items():
            mat2[k[0] - min_i, k[1] - min_j] = v

        max_x = max(k[0] for k in mat2)
        max_y = max(k[1] for k in mat2)

        mat = numpy.full((max_y + 1, max_x + 1), ' ')
        for k, v in mat2.items():
            mat[k[1], k[0]] = v
        rows = []
        for i in range(mat.shape[0]):
            rows.append(''.join(mat[i]))
        return "\n".join(rows)


class BiGraph:
    """
    BiGraph representation.

    .. versionadded:: 0.7
    """

    class A:
        "Additional information for a vertex or an edge."

        def __init__(self, kind):
            self.kind = kind

        def __repr__(self):
            return f"A({self.kind!r})"

    class B:
        "Additional information for a vertex or an edge."

        def __init__(self, name, content, onnx_name):
            if not isinstance(content, str):
                raise TypeError(  # pragma: no cover
                    f"content must be str not {type(content)!r}.")
            self.name = name
            self.content = content
            self.onnx_name = onnx_name

        def __repr__(self):
            return f"B({self.name!r}, {self.content!r}, {self.onnx_name!r})"

    def __init__(self, v0, v1, edges):
        """
        :param v0: first set of vertices (dictionary)
        :param v1: second set of vertices (dictionary)
        :param edges: edges
        """
        if not isinstance(v0, dict):
            raise TypeError("v0 must be a dictionary.")
        if not isinstance(v1, dict):
            raise TypeError("v0 must be a dictionary.")
        if not isinstance(edges, dict):
            raise TypeError("edges must be a dictionary.")
        self.v0 = v0
        self.v1 = v1
        self.edges = edges
        common = set(self.v0).intersection(set(self.v1))
        if len(common) > 0:
            raise ValueError(
                f"Sets v1 and v2 have common nodes (forbidden): {common!r}.")
        for a, b in edges:
            if a in v0 and b in v1:
                continue
            if a in v1 and b in v0:
                continue
            if b in v1:
                # One operator is missing one input.
                # We add one.
                self.v0[a] = BiGraph.A('ERROR')
                continue
            raise ValueError(
                f"Edges ({a!r}, {b!r}) not found among the vertices.")

    def __str__(self):
        """
        usual
        """
        return "%s(%d v., %d v., %d edges)" % (
            self.__class__.__name__, len(self.v0),
            len(self.v1), len(self.edges))

    def __iter__(self):
        """
        Iterates over all vertices and edges.
        It produces 3-uples:

        * 0, name, A: vertices in *v0*
        * 1, name, A: vertices in *v1*
        * -1, name, A: edges
        """
        for k, v in self.v0.items():
            yield 0, k, v
        for k, v in self.v1.items():
            yield 1, k, v
        for k, v in self.edges.items():
            yield -1, k, v

    def __getitem__(self, key):
        """
        Returns a vertex is key is a string or an edge
        if it is a tuple.

        :param key: vertex or edge name
        :return: value
        """
        if isinstance(key, tuple):
            return self.edges[key]
        if key in self.v0:
            return self.v0[key]
        return self.v1[key]

    def order_vertices(self):
        """
        Orders the vertices from the input to the output.

        :return: dictionary `{vertex name: order}`
        """
        order = {}
        for v in self.v0:
            order[v] = 0
        for v in self.v1:
            order[v] = 0
        modif = 1
        n_iter = 0
        while modif > 0:
            modif = 0
            for a, b in self.edges:
                if order[b] <= order[a]:
                    order[b] = order[a] + 1
                    modif += 1
            n_iter += 1
            if n_iter > len(order):
                break
        if modif > 0:
            raise RuntimeError(
                f"The graph has a cycle.\n{pprint.pformat(self.edges)}")
        return order

    def adjacency_matrix(self):
        """
        Builds an adjacency matrix.

        :return: matrix, list of row vertices, list of column vertices
        """
        order = self.order_vertices()
        ord_v0 = [(v, k) for k, v in order.items() if k in self.v0]
        ord_v1 = [(v, k) for k, v in order.items() if k in self.v1]
        ord_v0.sort()
        ord_v1.sort()
        row = [b for a, b in ord_v0]
        col = [b for a, b in ord_v1]
        row_id = {b: i for i, b in enumerate(row)}
        col_id = {b: i for i, b in enumerate(col)}
        matrix = numpy.zeros((len(row), len(col)), dtype=numpy.int32)
        for a, b in self.edges:
            if a in row_id:
                matrix[row_id[a], col_id[b]] = 1
            else:
                matrix[row_id[b], col_id[a]] = 1
        return matrix, row, col

    def display_structure(self, grid=5, distance=5):
        """
        Creates a display structure which contains
        all the necessary steps to display a graph.

        :param grid: align text to this grid
        :param distance: distance to the text
        :return: instance of @see cl AdjacencyGraphDisplay
        """
        def adjust(c, way):
            if way == 1:
                d = grid * ((c + distance * 2 - (grid // 2 + 1)) // grid)
            else:
                d = -grid * ((-c + distance * 2 - (grid // 2 + 1)) // grid)
            return d

        matrix, row, col = self.adjacency_matrix()
        row_id = {b: i for i, b in enumerate(row)}
        col_id = {b: i for i, b in enumerate(col)}

        interval_y_min = numpy.zeros((matrix.shape[0], ), dtype=numpy.int32)
        interval_y_max = numpy.zeros((matrix.shape[0], ), dtype=numpy.int32)
        interval_x_min = numpy.zeros((matrix.shape[1], ), dtype=numpy.int32)
        interval_x_max = numpy.zeros((matrix.shape[1], ), dtype=numpy.int32)
        interval_y_min[:] = max(matrix.shape)
        interval_x_min[:] = max(matrix.shape)

        graph = AdjacencyGraphDisplay()
        for key, value in self.edges.items():
            if key[0] in row_id:
                y = row_id[key[0]]
                x = col_id[key[1]]
            else:
                x = col_id[key[0]]
                y = row_id[key[1]]
            graph.add(x, y, 'cross', label=value.kind, orientation=(1, 0))
            if x < interval_y_min[y]:
                interval_y_min[y] = x
            if x > interval_y_max[y]:
                interval_y_max[y] = x
            if y < interval_x_min[x]:
                interval_x_min[x] = y
            if y > interval_x_max[x]:
                interval_x_max[x] = y

        for k, v in self.v0.items():
            y = row_id[k]
            x = adjust(interval_y_min[y], -1)
            graph.add(x, y, 'text', label=v.kind, orientation=(-1, 0))
            x = adjust(interval_y_max[y], 1)
            graph.add(x, y, 'text', label=k, orientation=(1, 0))

        for k, v in self.v1.items():
            x = col_id[k]
            y = adjust(interval_x_min[x], -1)
            graph.add(x, y, 'text', label=v.kind, orientation=(0, -1))
            y = adjust(interval_x_max[x], 1)
            graph.add(x, y, 'text', label=k, orientation=(0, 1))

        return graph

    def order(self):
        """
        Order nodes. Depth first.
        Returns a sequence of keys of mixed *v1*, *v2*.
        """
        # Creates forwards nodes.
        forwards = {}
        backwards = {}
        for k in self.v0:
            forwards[k] = []
            backwards[k] = []
        for k in self.v1:
            forwards[k] = []
            backwards[k] = []
        modif = True
        while modif:
            modif = False
            for edge in self.edges:
                a, b = edge
                if b not in forwards[a]:
                    forwards[a].append(b)
                    modif = True
                if a not in backwards[b]:
                    backwards[b].append(a)
                    modif = True

        # roots
        roots = [b for b, backs in backwards.items() if len(backs) == 0]
        if len(roots) == 0:
            raise RuntimeError(  # pragma: no cover
                "This graph has cycles. Not allowed.")

        # ordering
        order = {}
        stack = roots
        while len(stack) > 0:
            node = stack.pop()
            order[node] = len(order)
            w = forwards[node]
            if len(w) == 0:
                continue
            last = w.pop()
            stack.append(last)

        return order

    def summarize(self):
        """
        Creates a text summary of the graph.
        """
        order = self.order()
        keys = [(o, k) for k, o in order.items()]
        keys.sort()

        rows = []
        for _, k in keys:
            if k in self.v1:
                rows.append(str(self.v1[k]))
        return "\n".join(rows)

    @staticmethod
    def _onnx2bigraph_basic(model_onnx, recursive=False):
        """
        Implements graph type `'basic'` for function
        @see fn onnx2bigraph.
        """

        if recursive:
            raise NotImplementedError(  # pragma: no cover
                "Option recursive=True is not implemented yet.")
        v0 = {}
        v1 = {}
        edges = {}

        # inputs
        for i, o in enumerate(model_onnx.graph.input):
            v0[o.name] = BiGraph.A('Input-%d' % i)
        for i, o in enumerate(model_onnx.graph.output):
            v0[o.name] = BiGraph.A('Output-%d' % i)
        for o in model_onnx.graph.initializer:
            v0[o.name] = BiGraph.A('Init')
        for n in model_onnx.graph.node:
            nname = n.name if len(n.name) > 0 else "id%d" % id(n)
            v1[nname] = BiGraph.A(n.op_type)
            for i, o in enumerate(n.input):
                c = str(i) if i < 10 else "+"
                nname = n.name if len(n.name) > 0 else "id%d" % id(n)
                edges[o, nname] = BiGraph.A(f'I{c}')
            for i, o in enumerate(n.output):
                c = str(i) if i < 10 else "+"
                if o not in v0:
                    v0[o] = BiGraph.A('inout')
                nname = n.name if len(n.name) > 0 else "id%d" % id(n)
                edges[nname, o] = BiGraph.A(f'O{c}')

        return BiGraph(v0, v1, edges)

    @staticmethod
    def _onnx2bigraph_simplified(model_onnx, recursive=False):
        """
        Implements graph type `'simplified'` for function
        @see fn onnx2bigraph.
        """
        if recursive:
            raise NotImplementedError(  # pragma: no cover
                "Option recursive=True is not implemented yet.")
        v0 = {}
        v1 = {}
        edges = {}

        # inputs
        for o in model_onnx.graph.input:
            v0[f"I{len(v0)}"] = BiGraph.B(
                'In', make_hash_bytes(o.type.SerializeToString(), 2), o.name)
        for o in model_onnx.graph.output:
            v0[f"O{len(v0)}"] = BiGraph.B(
                'Ou', make_hash_bytes(o.type.SerializeToString(), 2), o.name)
        for o in model_onnx.graph.initializer:
            v0[f"C{len(v0)}"] = BiGraph.B(
                'Cs', make_hash_bytes(o.raw_data, 10), o.name)

        names_v0 = {v.onnx_name: k for k, v in v0.items()}

        for n in model_onnx.graph.node:
            key_node = f"N{len(v1)}"
            if len(n.attribute) > 0:
                ats = []
                for at in n.attribute:
                    ats.append(at.SerializeToString())
                ct = make_hash_bytes(b"".join(ats), 10)
            else:
                ct = ""
            v1[key_node] = BiGraph.B(
                n.op_type, ct, n.name)
            for o in n.input:
                key_in = names_v0[o]
                edges[key_in, key_node] = BiGraph.A('I')
            for o in n.output:
                if o not in names_v0:
                    key = f"R{len(v0)}"
                    v0[key] = BiGraph.B('Re', n.op_type, o)
                    names_v0[o] = key
                edges[key_node, key] = BiGraph.A('O')

        return BiGraph(v0, v1, edges)

    @staticmethod
    def onnx_graph_distance(onx1, onx2, verbose=0, fLOG=print):
        """
        Computes a distance between two ONNX graphs. They must not
        be too big otherwise this function might take for ever.
        The function relies on package :epkg:`mlstatpy`.

        :param onx1: first graph (ONNX graph or model file name)
        :param onx2: second graph (ONNX graph or model file name)
        :param verbose: verbosity
        :param fLOG: logging function
        :return: distance and differences

        .. warning::

            This is very experimental and very slow.

        .. versionadded:: 0.7
        """
        from mlstatpy.graph.graph_distance import GraphDistance

        if isinstance(onx1, str):
            onx1 = onnx.load(onx1)
        if isinstance(onx2, str):
            onx2 = onnx.load(onx2)

        def make_hash(init):
            return make_hash_bytes(init.raw_data)

        def build_graph(onx):
            edges = []
            labels = {}
            for node in onx.graph.node:
                if len(node.name) == 0:
                    name = str(id(node))
                else:
                    name = node.name
                for i in node.input:
                    edges.append((i, name))
                for p, i in enumerate(node.output):
                    edges.append((name, i))
                    labels[i] = "%s:%d" % (node.op_type, p)
                labels[name] = node.op_type
            for init in onx.graph.initializer:
                labels[init.name] = make_hash(init)

            g = GraphDistance(edges, vertex_label=labels)
            return g

        g1 = build_graph(onx1)
        g2 = build_graph(onx2)

        dist, gdist = g1.distance_matching_graphs_paths(
            g2, verbose=verbose, fLOG=fLOG, use_min=False)
        return dist, gdist


def onnx2bigraph(model_onnx, recursive=False, graph_type='basic'):
    """
    Converts an ONNX graph into a graph representation,
    edges, vertices.

    :param model_onnx: ONNX graph
    :param recursive: dig into subgraphs too
    :param graph_type: kind of graph it creates
    :return: see @cl BiGraph

    About *graph_type*:

    * `'basic'`: basic graph structure, it returns an instance
        of type @see cl BiGraph. The structure keeps the original
        names.
    * `'simplified'`: simplifed graph structure, names are removed
        as they could be prevent the algorithm to find any matching.

    .. exref::
        :title: Displays an ONNX graph as text

        The function uses an adjacency matrix of the graph.
        Results are displayed by rows, operator by columns.
        Results kinds are shows on the left,
        their names on the right. Operator types are displayed
        on the top, their names on the bottom.

        .. runpython::
            :showcode:

            import numpy
            from mlprodict.onnx_conv import to_onnx
            from mlprodict import __max_supported_opset__ as opv
            from mlprodict.tools.graphs import onnx2bigraph
            from mlprodict.npy.xop import loadop

            OnnxAdd, OnnxSub = loadop('Add', 'Sub')

            idi = numpy.identity(2).astype(numpy.float32)
            A = OnnxAdd('X', idi, op_version=opv)
            B = OnnxSub(A, 'W', output_names=['Y'], op_version=opv)
            onx = B.to_onnx({'X': idi, 'W': idi})
            bigraph = onnx2bigraph(onx)
            graph = bigraph.display_structure()
            text = graph.to_text()
            print(text)

    .. versionadded:: 0.7
    """
    if graph_type == 'basic':
        return BiGraph._onnx2bigraph_basic(
            model_onnx, recursive=recursive)
    if graph_type == 'simplified':
        return BiGraph._onnx2bigraph_simplified(
            model_onnx, recursive=recursive)
    raise ValueError(
        f"Unknown value for graph_type={graph_type!r}.")


def onnx_graph_distance(onx1, onx2, verbose=0, fLOG=print):
    """
    Computes a distance between two ONNX graphs. They must not
    be too big otherwise this function might take for ever.
    The function relies on package :epkg:`mlstatpy`.

    :param onx1: first graph (ONNX graph or model file name)
    :param onx2: second graph (ONNX graph or model file name)
    :param verbose: verbosity
    :param fLOG: logging function
    :return: distance and differences

    .. warning::

        This is very experimental and very slow.

    .. versionadded:: 0.7
    """
    return BiGraph.onnx_graph_distance(onx1, onx2, verbose=verbose, fLOG=fLOG)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from ._op import OpRunUnaryNum
from ._new_ops import OperatorSchema


class YieldOp(OpRunUnaryNum):

    atts = {'full_shape_outputs': [],
            'non_differentiable_outputs': []}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               expected_attributes=YieldOp.atts,
                               **options)

    def _find_custom_operator_schema(self, op_name):
        if op_name == "YieldOp":
            return YieldOpSchema()
        raise RuntimeError(  # pragma: no cover
            f"Unable to find a schema for operator '{op_name}'.")

    def _run(self, a, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.inplaces.get(0, False):
            return (a, )
        return (a.copy(), )

    def to_python(self, inputs):
        return "", f"return {inputs[0]}.copy()"


class YieldOpSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl ComplexAbs.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'YieldOp')
        self.attributes = {}



"""
@file
@brief A couple of tools unrelated to what the package does.
"""
import pickle
import re
import types
import numpy


def numpy_min_max(x, fct, minmax=False):
    """
    Returns the minimum of an array.
    Deals with text as well.
    """
    try:
        if hasattr(x, 'todense'):
            x = x.todense()
        if (x.dtype.kind[0] not in 'Uc' or
                x.dtype in {numpy.uint8}):
            return fct(x)
        try:  # pragma: no cover
            x = x.ravel()
        except AttributeError:  # pragma: no cover
            pass
        keep = list(filter(lambda s: isinstance(s, str), x))
        if len(keep) == 0:  # pragma: no cover
            return numpy.nan
        keep.sort(reverse=minmax)
        val = keep[0]
        if len(val) > 10:  # pragma: no cover
            val = val[:10] + '...'
        return f"{val!r}"
    except (ValueError, TypeError, AttributeError):
        return '?'


def numpy_min(x):
    """
    Returns the maximum of an array.
    Deals with text as well.
    """
    return numpy_min_max(x, lambda x: x.min(), minmax=False)


def numpy_max(x):
    """
    Returns the maximum of an array.
    Deals with text as well.
    """
    return numpy_min_max(x, lambda x: x.max(), minmax=True)


def debug_dump(clname, obj, folder=None, ops=None):
    """
    Dumps an object for debug purpose.

    @param      clname  class name
    @param      obj     object
    @param      folder  folder
    @param      ops     operator to dump
    @return             filename
    """
    def debug_print_(obj, prefix=''):
        name = clname
        if isinstance(obj, dict):
            if 'in' in obj and 'out' in obj:
                nan_in = any(map(lambda o: any(map(numpy.isnan, o.ravel())),
                                 obj['in']))
                nan_out = any(map(lambda o: any(map(numpy.isnan, o.ravel())),
                                  obj['out']))
                if not nan_in and nan_out:
                    print("NAN-notin-out ", name, prefix,
                          {k: getattr(ops, k, '?') for k in getattr(ops, 'atts', {})})
                    return True
                return False  # pragma: no cover
            for k, v in obj.items():  # pragma: no cover
                debug_print_([v], k)
            return None  # pragma: no cover
        if isinstance(obj, list):
            for i, o in enumerate(obj):
                if o is None:
                    continue
                if any(map(numpy.isnan, o.ravel())):
                    print("NAN", prefix, i, name, o.shape)
            return None
        raise NotImplementedError(  # pragma: no cover
            f"Unable to debug object of type {type(obj)}.")

    dump = debug_print_(obj)
    if dump:
        name = f'cpu-{clname}-{id(obj)}-{id(ops)}.pkl'
        if folder is not None:
            name = "/".join([folder, name])
        with open(name, 'wb') as f:
            pickle.dump(obj, f)
        return name
    return None


def debug_print(k, obj, printed):
    """
    Displays informations on an object.

    @param      k       name
    @param      obj     object
    @param      printed memorizes already printed object
    """
    if k not in printed:
        printed[k] = obj
        if hasattr(obj, 'shape'):
            print("-='{}' shape={} dtype={} min={} max={}{}".format(
                  k, obj.shape, obj.dtype, numpy_min(obj),
                  numpy_max(obj),
                  ' (sparse)' if 'coo_matrix' in str(type(obj)) else ''))
        elif (isinstance(obj, list) and len(obj) > 0 and
                not isinstance(obj[0], dict)):  # pragma: no cover
            print(f"-='{k}' list len={len(obj)} min={min(obj)} max={max(obj)}")
        else:  # pragma: no cover
            print(f"-='{k}' type={type(obj)}")


def make_callable(fct, obj, code, gl, debug):
    """
    Creates a callable function able to
    cope with default values as the combination
    of functions *compile* and *exec* does not seem
    able to take them into account.

    @param      fct     function name
    @param      obj     output of function *compile*
    @param      code    code including the signature
    @param      gl      context (local and global)
    @param      debug   add debug function
    @return             callable functions
    """
    cst = "def " + fct + "("
    sig = None
    for line in code.split('\n'):
        if line.startswith(cst):
            sig = line
            break
    if sig is None:  # pragma: no cover
        raise ValueError(
            f"Unable to find function '{fct}' in\n{code}")
    reg = re.compile(
        "([a-z][A-Za-z_0-9]*)=((None)|(False)|(True)|([0-9.e+-]+))")
    fall = reg.findall(sig)
    defs = []
    for name_value in fall:
        name = name_value[0]
        value = name_value[1]
        if value == 'None':
            defs.append((name, None))
            continue
        if value == 'True':
            defs.append((name, True))
            continue
        if value == 'False':
            defs.append((name, False))
            continue
        f = float(value)
        if int(f) == f:
            f = int(f)
        defs.append((name, f))

    # debug
    if debug:
        gl = gl.copy()
        gl['debug_print'] = debug_print
        gl['print'] = print
    # specific
    if "value=array([0.], dtype=float32)" in sig:
        defs.append(('value', numpy.array([0.], dtype=numpy.float32)))
    res = types.FunctionType(obj, gl, fct, tuple(_[1] for _ in defs))
    if res.__defaults__ != tuple(_[1] for _ in defs):  # pylint: disable=E1101
        # See https://docs.python.org/3/library/inspect.html
        # See https://stackoverflow.com/questions/11291242/python-dynamically-create-function-at-runtime
        lines = [str(sig)]  # pragma: no cover
        for name in ['co_argcount', 'co_cellvars', 'co_code', 'co_consts', 'co_filename',
                     'co_firstlineno', 'co_flags', 'co_freevars', 'co_kwonlyargcount',
                     'co_lnotab', 'co_name', 'co_names', 'co_nlocals', 'co_stacksize',
                     'co_varnames']:  # pragma: no cover
            v = getattr(res.__code__, name, None)  # pylint: disable=E1101
            if v is not None:
                lines.append(f'{name}={v!r}')
        raise RuntimeError(  # pragma: no cover
            "Defaults values of function '{}' (defaults={}) are missing.\nDefault: "
            "{}\n{}\n----\n{}".format(
                fct, res.__defaults__, defs, "\n".join(lines), code))  # pylint: disable=E1101
    return res


def print_code(code, begin=1):
    """
    Returns the code with line number.
    """
    rows = code.split("\n")
    return "\n".join("%03d %s" % (i + begin, s)
                     for i, s in enumerate(rows))



"""
@file
@brief Runtime to infer shapes.

.. versionadded:: 0.9
"""
import numpy
from onnx import FunctionProto, ModelProto
from onnx.numpy_helper import to_array
from onnx.mapping import TENSOR_TYPE_TO_NP_TYPE
from .ops_shape.shape_result import ShapeResult
from .ops_shape.shape_container import ShapeContainer
from .ops_shape import shape_dispatch


class OnnxShapeInference:
    """
    Implements a micro runtime for ONNX graphs.
    It does not implements all the operator types.

    :param model_onnx: ONNX model

    Other attributes:

    * `known_shapes_`: shapes which can be inferred without any input
    * `cache_`: keeps track of the function used to infer
      the shapes
    * `is_isfunction`: tells if the graph is a function or a model

    .. runpython::
        :showcode:

        import pprint
        import numpy
        from mlprodict.onnxrt.onnx_shape_inference import OnnxShapeInference
        from mlprodict.npy.xop_variable import Variable
        from mlprodict.npy.xop import loadop

        opset = 15
        OnnxAdd = loadop('Add')
        dtype = numpy.float32

        cop = OnnxAdd('X', numpy.array(
            [[1]], dtype=dtype), op_version=opset)
        cop4 = OnnxAdd(cop, numpy.array([[2]], dtype=dtype),
                       output_names=['Y'])
        vari = Variable('X', numpy.float32, [None, 3])
        model_def = cop4.to_onnx([vari], run_shape=False)
        rt = OnnxShapeInference(model_def)
        out = rt.run()
        pprint.pprint(out.get())
    """

    def __init__(self, model_onnx):
        if not isinstance(model_onnx, (FunctionProto, ModelProto)):
            raise TypeError(  # pragma: no cover
                "model_onnx is not from FunctionProto or ModelProto "
                "%r." % type(model_onnx))
        self.is_function = isinstance(model_onnx, FunctionProto)
        self.model_onnx = model_onnx
        self.cache_ = {}
        self.known_shapes_ = self._run_empty()

    @property
    def input_names(self):
        "Returns input names."
        if self.is_function:
            return list(self.model_onnx.input)
        return [i.name for i in self.model_onnx.graph.input]

    @property
    def output_names(self):
        "Returns output names."
        if self.is_function:
            return list(self.model_onnx.output)
        return [i.name for i in self.model_onnx.graph.output]

    def __repr__(self):
        "Usual"
        return f"{self.__class__.__name__}(...)"

    @staticmethod
    def _get_shape(obj, known_shapes=None, result_name=None):
        if obj is None:
            return [], None, False
        dtype = TENSOR_TYPE_TO_NP_TYPE.get(
            obj.type.tensor_type.elem_type, None)
        shape = []
        for dimi, d in enumerate(obj.type.tensor_type.shape.dim):
            v = d.dim_value if d.dim_value > 0 else d.dim_param
            if v in ('', None):
                if known_shapes is None or result_name is None:
                    raise RuntimeError(  # pragma: no cover
                        "known_shapes must be specified if "
                        "a dimension is not.")
                v = known_shapes.get_new_name(v, result_name, dimi)
            shape.append(v)
        return shape, dtype, False

    def _run_empty(self):
        """
        Computes shape and types of all results.

        :return: all intermediates results and output as a dictionary
        """
        def get_obj(name, inputs):
            if self.is_function:
                return None
            if inputs:
                for o in self.model_onnx.graph.input:
                    if o.name == name:
                        return o
            else:
                for o in self.model_onnx.graph.output:
                    if o.name == name:
                        return o
            return None

        known_shapes = ShapeContainer()
        if not self.is_function:
            for init in self.model_onnx.graph.initializer:
                mat = to_array(init)
                known_shapes.update(init.name, ShapeResult(
                    init.name, mat.shape, mat.dtype, sparse=False))

        for name in self.input_names:
            if name in known_shapes:
                raise NotImplementedError(
                    f"Optional inputs are not implemented yet. (name={name!r})")
            shape, dtype, sparse = self._get_shape(
                get_obj(name, True), known_shapes, result_name=name)
            known_shapes.update(name, ShapeResult(
                name, shape, dtype, sparse=sparse))

        for name in self.output_names:
            if name in known_shapes:
                raise NameError(  # pragma: no cover
                    f"Output {name!r} is already present. Use Identity node.")
            shape, dtype, sparse = self._get_shape(
                get_obj(name, False), known_shapes, result_name=name)
            if dtype is None:
                # The onnx graph was created with named outputs
                # but with no type or shape.
                continue
            known_shapes.update(name, ShapeResult(
                name, shape, dtype, sparse=sparse))

        nodes = (
            self.model_onnx.node if self.is_function
            else self.model_onnx.graph.node)
        cont = True
        while cont:
            cont = False
            for node in nodes:
                cont = cont or shape_dispatch(
                    self.cache_, known_shapes, node, rt_class=self.__class__)
        return known_shapes

    def run(self, inputs=None):
        """
        Runs shape inference and type given known inputs.

        :param inputs: inputs
        :return: all results
        """
        known_shapes = self.known_shapes_.copy(deep=True)
        if inputs is None:
            known_shapes.resolve()
            return known_shapes

        cont = False
        for name, obj in inputs.items():
            shape, dtype, sparse = (
                obj.shape, obj.dtype, not isinstance(obj, numpy.ndarray))
            cont = cont or known_shapes.update(
                name, ShapeResult(name, shape, dtype, sparse=sparse))

        nodes = (
            self.model_onnx.node if self.is_function
            else self.model_onnx.graph.node)
        while cont:
            cont = False
            for node in nodes:
                updated = shape_dispatch(
                    self.cache_, known_shapes, node, rt_class=self.__class__)
                cont = cont or updated
        known_shapes.resolve()
        return known_shapes



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def common_reference_implementation(data, shape):
    ones = numpy.ones(shape, dtype=data.dtype)
    return data * ones


class CommonExpand(OpRun):

    def __init__(self, onnx_node, desc=None, expected_attributes=None, **options):
        OpRun.__init__(
            self, onnx_node, desc=desc,
            expected_attributes=expected_attributes, **options)

    def _run(self, data, shape, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return (common_reference_implementation(data, shape), )


class Expand_13(CommonExpand):

    def __init__(self, onnx_node, desc=None, **options):
        CommonExpand.__init__(
            self, onnx_node, desc=desc, **options)


Expand = Expand_13



"""
@file
@brief Implementation of a dummy score using
:epkg:`cdist`.
"""
import numpy
from onnx import onnx_pb as onnx_proto
from scipy.spatial.distance import cdist


def score_cdist_sum(expected_values, predictions,
                    metric='sqeuclidean', p=None):
    """
    Computes the sum of pairwise distances between
    *expected_values* and *predictions*. It has no
    particular purpose except the one of converting
    a scorer into ONNX.

    @param      expected_values     expected_values
    @param      predictions         predictions
    @param      metric              see function :epkg:`cdist`
    @param      p                   see function :epkg:`cdist`
    @return                         some of the pairwise distances
    """
    if p is None:
        dist = cdist(expected_values, predictions, metric=metric)
    else:
        dist = cdist(expected_values, predictions, metric=metric, p=p)
    return numpy.sum(dist, axis=1)


def convert_score_cdist_sum(scope, operator, container):
    """
    Converts function @see fn score_cdist_sum into :epkg:`ONNX`.
    """
    op = operator.raw_operator
    if op._fct != score_cdist_sum:  # pylint: disable=W0143
        raise RuntimeError(  # pragma: no cover
            f"The wrong converter was called {op._fct} != {score_cdist_sum}.")

    from skl2onnx.algebra.complex_functions import onnx_cdist
    from skl2onnx.algebra.onnx_ops import OnnxReduceSumApi11  # pylint: disable=E0611
    from skl2onnx.common.data_types import guess_numpy_type

    X = operator.inputs[0]
    Y = operator.inputs[1]
    out = operator.outputs
    opv = container.target_opset
    dtype = guess_numpy_type(operator.inputs[0].type)
    out = operator.outputs

    options = container.get_options(score_cdist_sum, dict(cdist=None))

    kwargs = op.kwargs

    if options.get('cdist', None) == 'single-node':
        attrs = kwargs
        cdist_name = scope.get_unique_variable_name('cdist')
        container.add_node('CDist', [X.full_name, Y.full_name], cdist_name,
                           op_domain='mlprodict', name=scope.get_unique_operator_name('CDist'),
                           **attrs)
        if container.target_opset < 13:
            container.add_node('ReduceSum', [cdist_name], out[0].full_name,
                               axes=[1], keepdims=0,
                               name=scope.get_unique_operator_name('ReduceSum'))
        else:
            axis_name = scope.get_unique_variable_name('axis')
            container.add_initializer(
                axis_name, onnx_proto.TensorProto.INT64, [1], [1])  # pylint: disable=E1101
            container.add_node(
                'ReduceSum', [cdist_name, axis_name],
                out[0].full_name, keepdims=0,
                name=scope.get_unique_operator_name('ReduceSum'))
    else:
        metric = kwargs['metric']
        if metric == 'minkowski':
            dists = onnx_cdist(X, Y, dtype=dtype, op_version=opv,
                               metric=metric, p=kwargs.get('p', 2))
        else:
            dists = onnx_cdist(X, Y, dtype=dtype, op_version=opv,
                               metric=kwargs['metric'])

        res = OnnxReduceSumApi11(dists, axes=[1], keepdims=0,
                                 output_names=[out[0].full_name],
                                 op_version=opv)
        res.add_to(scope, container)



# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *ops_cpu*.
"""
from ._op import OpRunOnnxRuntime


def load_op(onnx_node, desc=None, options=None, variables=None,
            dtype=None, runtime=None):
    """
    Gets the operator related to the *onnx* node.

    :param onnx_node: :epkg:`onnx` node
    :param desc: internal representation
    :param options: runtime options
    :param variables: registered variables created by previous operators
    :param dtype: float computation type
    :param runtime: runtime
    :return: runtime class
    """
    if desc is None:
        raise ValueError(  # pragma: no cover
            "desc should not be None.")
    return OpRunOnnxRuntime(onnx_node, desc, variables=variables,
                            dtype=dtype, runtime=runtime, **options)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from ...onnx_tools.onnx2py_helper import guess_dtype
from ._op import OpRun
from ._new_ops import OperatorSchema


class Expression(OpRun):

    atts = {
        'expression': None,
    }

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Expression.atts,
                       **options)
        if not hasattr(self.expression, 'run'):
            raise RuntimeError(  # pragma: no cover
                "Parameter 'expression' must have a method 'run', "
                "type {}.".format(type(self.then_branch)))

        self._run_expression = (self.expression.run_in_scan
                                if hasattr(self.expression, 'run_in_scan')
                                else self.expression.run)
        self.additional_inputs = list(self.expression.static_inputs)
        self.input_names = [
            i.name for i in self.onnx_node.attribute[0].g.input]

    def _find_custom_operator_schema(self, op_name):
        if op_name == "Expression":
            return ExpressionSchema()
        raise RuntimeError(  # pragma: no cover
            f"Unable to find a schema for operator '{op_name}'.")

    def need_context(self):
        """
        Tells the runtime if this node needs the context
        (all the results produced so far) as it may silently access
        one of them (operator Loop).
        The default answer is `False`.
        """
        return True

    def _run(self, *inputs, named_inputs=None, context=None,  # pylint: disable=W0221
             attributes=None, verbose=0, fLOG=None):

        if verbose > 0 and fLOG is not None:
            fLOG(f'  -- expression> {list(context)!r}')
        if named_inputs is None:
            if len(inputs) != len(self.input_names):
                raise RuntimeError(  # pragma: no cover
                    "Unpexpected number of inputs (%d != %d): %r." % (
                        len(inputs), len(self.input_names), self.input_names))
            named_inputs = {name: value for name,
                            value in zip(self.input_names, inputs)}
        outputs = self._run_expression(named_inputs, context=context,
                                       attributes=attributes,
                                       verbose=verbose, fLOG=fLOG)
        if verbose > 0 and fLOG is not None:
            fLOG('  -- expression<')
        final = tuple([outputs[name]
                      for name in self.expression.output_names])
        return final

    def _pick_type(self, res, name):
        if name in res:
            return res[name]
        out = {o.name: o for o in self.expression.obj.graph.output}
        if name not in out:
            raise ValueError(
                "Unable to find name=%r in %r or %r." % (
                    name, list(sorted(res)), list(sorted(out))))
        dt = out[name].type.tensor_type.elem_type
        return guess_dtype(dt)


class ExpressionSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl ComplexAbs.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'Expression')
        self.attributes = Expression.atts



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def _apply_adagrad(r, t, x, g, h, norm_coefficient,
                   epsilon, decay_factor):
    # Compute adjusted learning-rate.
    r_ = r / (1 + t * decay_factor)
    # Add gradient of regularization term.
    g_regularized = norm_coefficient * x + g
    # Update squared accumulated gradient.
    h_new = h + g_regularized * g_regularized
    # Compute ADAGRAD's gradient scaling factors
    h_sqrt = numpy.sqrt(h_new) + epsilon
    # Apply ADAGRAD update rule.
    x_new = x - r_ * g_regularized / h_sqrt
    return (x_new, h_new)


class Adagrad(OpRun):

    atts = {'decay_factor': 0.,
            'epsilon': 9.999999974752427e-07,
            'norm_coefficient': 0.}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Adagrad.atts,
                       **options)

    def _run(self, *data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if len(data) == 5:
            return self._run1(*data)
        n = (len(data) - 2) // 3
        xs = []
        hs = []
        for i in range(0, n):
            a, b = self._run1(*data[:2], data[2 + i],
                              data[2 + n + i], data[2 + n * 2 + i])
            xs.append(a)
            hs.append(b)
        return tuple(xs + hs)

    def _run1(self, r, t, x, g, h):  # pylint: disable=W0221
        x_new, h_new = _apply_adagrad(
            r, t, x, g, h, self.norm_coefficient, self.epsilon, self.decay_factor)
        return x_new, h_new



"""
@file
@brief Useful plots.
"""
import numpy


def heatmap(data, row_labels, col_labels, ax=None,
            cbar_kw=None, cbarlabel=None, **kwargs):
    """
    Creates a heatmap from a numpy array and two lists of labels.
    See @see fn plot_benchmark_metrics for an example.

    @param  data        a 2D numpy array of shape (N, M).
    @param  row_labels  a list or array of length N with the labels for the rows.
    @param  col_labels  a list or array of length M with the labels for the columns.
    @param  ax          a `matplotlib.axes.Axes` instance to which the heatmap is plotted,
                        if not provided, use current axes or create a new one. Optional.
    @param  cbar_kw     a dictionary with arguments to `matplotlib.Figure.colorbar
                        <https://matplotlib.org/api/_as_gen/matplotlib.pyplot.colorbar.html>`_.
                        Optional.
    @param  cbarlabel   the label for the colorbar. Optional.
    @param   kwargs     all other arguments are forwarded to `imshow
                        <https://matplotlib.org/api/_as_gen/matplotlib.pyplot.imshow.html>`_
    @return             ax, image, color bar
    """
    import matplotlib.pyplot as plt  # delayed

    if not ax:
        ax = plt.gca()  # pragma: no cover

    # Plot the heatmap
    im = ax.imshow(data, **kwargs)

    # Create colorbar
    if cbar_kw is None:
        cbar_kw = {}
    cbar = ax.figure.colorbar(im, ax=ax, **cbar_kw)
    if cbarlabel is not None:
        cbar.ax.set_ylabel(cbarlabel, rotation=-90, va="bottom")

    # We want to show all ticks...
    ax.set_xticks(numpy.arange(data.shape[1]))
    ax.set_yticks(numpy.arange(data.shape[0]))
    # ... and label them with the respective list entries.
    ax.set_xticklabels(col_labels)
    ax.set_yticklabels(row_labels)

    # Let the horizontal axes labeling appear on top.
    ax.tick_params(top=True, bottom=False,
                   labeltop=True, labelbottom=False)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=-30, ha="right",
             rotation_mode="anchor")

    # Turn spines off and create white grid.
    for _, spine in ax.spines.items():
        spine.set_visible(False)

    ax.set_xticks(numpy.arange(data.shape[1] + 1) - .5, minor=True)
    ax.set_yticks(numpy.arange(data.shape[0] + 1) - .5, minor=True)
    ax.grid(which="minor", color="w", linestyle='-', linewidth=3)
    ax.tick_params(which="minor", bottom=False, left=False)
    return ax, im, cbar


def annotate_heatmap(im, data=None, valfmt="{x:.2f}",
                     textcolors=("black", "black"),
                     threshold=None, **textkw):
    """
    Annotates a heatmap.
    See @see fn plot_benchmark_metrics for an example.

    @param  im          the *AxesImage* to be labeled.
    @param  data        data used to annotate. If None, the image's data is used. Optional.
    @param  valfmt      the format of the annotations inside the heatmap. This should either
                        use the string format method, e.g. `"$ {x:.2f}"`, or be a
                        `matplotlib.ticker.Formatter
                        <https://matplotlib.org/api/ticker_api.html>`_. Optional.
    @param  textcolors  a list or array of two color specifications.  The first is used for
                        values below a threshold, the second for those above. Optional.
    @param  threshold   value in data units according to which the colors from textcolors are
                        applied. If None (the default) uses the middle of the colormap as
                        separation. Optional.
    @param  textkw      all other arguments are forwarded to each call to `text` used to create
                        the text labels.
    @return             annotated objects
    """
    if not isinstance(data, (list, numpy.ndarray)):
        data = im.get_array()
    if threshold is not None:
        threshold = im.norm(threshold)  # pragma: no cover
    else:
        threshold = im.norm(data.max()) / 2.

    kw = dict(horizontalalignment="center", verticalalignment="center")
    kw.update(textkw)

    # Get the formatter in case a string is supplied
    if isinstance(valfmt, str):
        import matplotlib  # delayed
        valfmt = matplotlib.ticker.StrMethodFormatter(valfmt)

    texts = []
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            kw.update(color=textcolors[int(im.norm(data[i, j]) > threshold)])
            text = im.axes.text(j, i, valfmt(data[i, j], None), **kw)
            texts.append(text)

    return texts


def plot_benchmark_metrics(metric, xlabel=None, ylabel=None,
                           middle=1., transpose=False, ax=None,
                           cbar_kw=None, cbarlabel=None,
                           valfmt="{x:.2f}x"):
    """
    Plots a heatmap which represents a benchmark.
    See example below.

    @param      metric      dictionary ``{ (x,y): value }``
    @param      xlabel      x label
    @param      ylabel      y label
    @param      middle      force the white color to be this value
    @param      transpose   switches *x* and *y*
    @param      ax          axis to borrow
    @param      cbar_kw     a dictionary with arguments to `matplotlib.Figure.colorbar
                            <https://matplotlib.org/api/_as_gen/matplotlib.pyplot.colorbar.html>`_.
                            Optional.
    @param      cbarlabel   the label for the colorbar. Optional.
    @param      valfmt      format for the annotations
    @return                 ax, colorbar

    .. exref::
        :title: Plot benchmark improvments
        :lid: plot-2d-benchmark-metric

        .. plot::

            import matplotlib.pyplot as plt
            from mlprodict.plotting.plotting_benchmark import plot_benchmark_metrics

            data = {(1, 1): 0.1, (10, 1): 1, (1, 10): 2,
                    (10, 10): 100, (100, 1): 100, (100, 10): 1000}

            fig, ax = plt.subplots(1, 2, figsize=(10, 4))
            plot_benchmark_metrics(data, ax=ax[0], cbar_kw={'shrink': 0.6})
            plot_benchmark_metrics(data, ax=ax[1], transpose=True,
                                   xlabel='X', ylabel='Y',
                                   cbarlabel="ratio")
            plt.show()
    """
    if transpose:
        metric = {(k[1], k[0]): v for k, v in metric.items()}
        return plot_benchmark_metrics(metric, ax=ax, xlabel=ylabel, ylabel=xlabel,
                                      middle=middle, transpose=False,
                                      cbar_kw=cbar_kw, cbarlabel=cbarlabel)

    from matplotlib.colors import LogNorm  # delayed

    x = numpy.array(list(sorted(set(k[0] for k in metric))))
    y = numpy.array(list(sorted(set(k[1] for k in metric))))
    rx = {v: i for i, v in enumerate(x)}
    ry = {v: i for i, v in enumerate(y)}

    X, _ = numpy.meshgrid(x, y)
    zm = numpy.zeros(X.shape, dtype=numpy.float64)
    for k, v in metric.items():
        zm[ry[k[1]], rx[k[0]]] = v

    xs = [str(_) for _ in x]
    ys = [str(_) for _ in y]
    vmin = min(metric.values())
    vmax = max(metric.values())
    if middle is not None:
        v1 = middle / vmin
        v2 = middle / vmax
        vmin = min(vmin, v2)
        vmax = max(vmax, v1)
    ax, im, cbar = heatmap(zm, ys, xs, ax=ax, cmap="bwr",
                           norm=LogNorm(vmin=vmin, vmax=vmax),
                           cbarlabel=cbarlabel, cbar_kw=cbar_kw)
    annotate_heatmap(im, valfmt=valfmt)
    if xlabel is not None:
        ax.set_xlabel(xlabel)
    if ylabel is not None:
        ax.set_ylabel(ylabel)
    return ax, cbar



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.

.. versionadded:: 0.7
"""
import numpy
from ._op import OpRun


class Loop(OpRun):

    atts = {'body': None}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Loop.atts,
                       **options)
        if not hasattr(self.body, 'run'):
            raise RuntimeError(  # pragma: no cover
                f"Parameter 'body' must have a method 'run', type {type(self.body)}.")

        self._run_meth = (self.body.run_in_scan
                          if hasattr(self.body, 'run_in_scan')
                          else self.body.run)
        self.additional_inputs = self.body.static_inputs

    def need_context(self):
        """
        The operator Loop needs to know all results produced
        so far as the loop may silently access one of them.
        Some information are not always referred in the list of inputs
        (kind of static variables).
        """
        return len(self.additional_inputs) > 0

    def _run(self, M, cond, v_initial, *args, callback=None, context=None,  # pylint: disable=W0221
             attributes=None, verbose=0, fLOG=None):
        loop_inputs = self.body.input_names
        inputs = {name: None for name in loop_inputs}
        inputs[loop_inputs[2]] = v_initial
        cond_name = self.body.output_names[0]
        if len(args) > 0:
            begin = len(loop_inputs) - len(args)
            all_inputs = loop_inputs[begin:]
            for name, val in zip(all_inputs, args):
                inputs[name] = val
        if len(self.additional_inputs) > 0:
            if context is None:
                raise RuntimeError(
                    "Additional inputs %r are missing and context is None."
                    "" % (self.additional_inputs, ))
            for a in self.additional_inputs:
                if a in context:
                    inputs[a] = context[a]
                else:
                    raise RuntimeError(
                        "Additional inputs %r not found in context\n%s." % (
                            a, "\n".join(sorted(map(str, context)))))

        it = 0
        while cond and it < M:
            inputs[self.body.input_names[0]] = numpy.array(it, dtype=M.dtype)
            inputs[self.body.input_names[1]] = cond
            outputs = self._run_meth(inputs)
            cond = outputs[cond_name]
            if cond is None:
                raise RuntimeError(
                    f"condition {cond_name!r} returned by the subgraph cannot be None.")
            for i, o in zip(self.body.input_names[2:],
                            self.body.output_names[1:]):
                inputs[i] = outputs[o]
            if callback is not None:
                callback(inputs, context=context)
            it += 1

        if it == 0:
            outputs = {self.body.output_names[1]: cond}
            for i, o in zip(self.body.input_names[2:],
                            self.body.output_names[1:]):
                outputs[o] = inputs[i]
        for o in self.body.output_names:
            if o not in outputs:
                outputs[o] = numpy.empty(shape=tuple())
        res = tuple([outputs[name] for name in self.body.output_names[1:]])
        if any(r is None for r in res):
            raise TypeError(  # pragma: no cover
                "Operator Loop produces a None value.")
        return res



# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *onnxrt*.
"""
from .onnx_inference import OnnxInference
from .onnx_micro_runtime import OnnxMicroRuntime
from .onnx_shape_inference import OnnxShapeInference



"""
@file
@brief Command line about validation of prediction runtime.
"""
import os
from io import StringIO
from logging import getLogger
import warnings
import json
from multiprocessing import Pool
from pandas import DataFrame, read_csv, concat
from sklearn.exceptions import ConvergenceWarning


def benchmark_doc(runtime, black_list=None, white_list=None,
                  out_raw='bench_raw.xlsx', out_summary="bench_summary.xlsx",
                  dump_dir='dump', fLOG=print, verbose=0):
    """
    Runs the benchmark published into the documentation
    (see :ref:`l-onnx-bench-onnxruntime1` and
    :ref:`l-onnx-bench-python_compiled`).

    :param runtime: runtime (python, python_compiled,
        onnxruntime1, onnxruntime2)
    :param black_list: models to skip, None for none
        (comma separated list)
    :param white_list: models to benchmark, None for all
        (comma separated list)
    :param out_raw: all results are saved in that file
    :param out_summary: all results are summarized in that file
    :param dump_dir: folder where to dump intermediate results
    :param fLOG: logging function
    :param verbose: verbosity
    :return: list of created files
    """
    def _save(df, name):
        ext = os.path.splitext(name)[-1]
        if ext == '.xlsx':
            df.to_excel(name, index=False)
        elif ext == '.csv':
            df.to_csv(name, index=False)
        else:
            raise ValueError(  # pragma: no cover
                f"Unexpected extension in {name!r}.")
        if verbose > 1:
            fLOG(  # pragma: no cover
                f"[mlprodict] wrote '{name}'")

    from pyquickhelper.loghelper import run_cmd
    from pyquickhelper.loghelper.run_cmd import get_interpreter_path
    from tqdm import tqdm
    from ..onnxrt.validate.validate_helper import sklearn_operators
    from ..onnx_conv import register_converters, register_rewritten_operators
    register_converters()
    try:
        register_rewritten_operators()
    except KeyError:  # pragma: no cover
        warnings.warn("converter for HistGradientBoosting* not not exist. "
                      "Upgrade sklearn-onnx")

    if black_list is None:
        black_list = []
    else:
        black_list = black_list.split(',')
    if white_list is None:
        white_list = []
    else:
        white_list = white_list.split(',')

    filenames = []
    skls = sklearn_operators(extended=True)
    skls = [_['name'] for _ in skls]
    if white_list:
        skls = [_ for _ in skls if _ in white_list]
    skls.sort()
    if verbose > 0:
        pbar = tqdm(skls)
    else:
        pbar = skls
    for op in pbar:
        if black_list is not None and op in black_list:
            continue
        if verbose > 0:
            pbar.set_description(  # pragma: no cover
                f"[{op + ' ' * (25 - len(op))}]")

        loop_out_raw = os.path.join(
            dump_dir, f"bench_raw_{runtime}_{op}.csv")
        loop_out_sum = os.path.join(
            dump_dir, f"bench_sum_{runtime}_{op}.csv")
        cmd = ('{0} -m mlprodict validate_runtime --verbose=0 --out_raw={1} --out_summary={2} '
               '--benchmark=1 --dump_folder={3} --runtime={4} --models={5}'.format(
                   get_interpreter_path(), loop_out_raw, loop_out_sum, dump_dir, runtime, op))
        if verbose > 1:
            fLOG(f"[mlprodict] cmd '{cmd}'.")  # pragma: no cover
        out, err = run_cmd(cmd, wait=True, fLOG=None)
        if not os.path.exists(loop_out_sum):  # pragma: no cover
            if verbose > 2:
                fLOG(f"[mlprodict] unable to find '{loop_out_sum}'.")
            if verbose > 1:
                fLOG(f"[mlprodict] cmd '{cmd}'")
                fLOG(f"[mlprodict] unable to find '{loop_out_sum}'")
            msg = "Unable to find '{}'\n--CMD--\n{}\n--OUT--\n{}\n--ERR--\n{}".format(
                loop_out_sum, cmd, out, err)
            if verbose > 1:
                fLOG(msg)
            rows = [{'name': op, 'scenario': 'CRASH',
                     'ERROR-msg': msg.replace("\n", " -- ")}]
            df = DataFrame(rows)
            df.to_csv(loop_out_sum, index=False)
        filenames.append((loop_out_raw, loop_out_sum))

    # concatenate summaries
    dfs_raw = [read_csv(name[0])
               for name in filenames if os.path.exists(name[0])]
    dfs_sum = [read_csv(name[1])
               for name in filenames if os.path.exists(name[1])]
    df_raw = concat(dfs_raw, sort=False)
    piv = concat(dfs_sum, sort=False)

    opset_cols = [(int(oc.replace("opset", "")), oc)
                  for oc in piv.columns if 'opset' in oc]
    opset_cols.sort(reverse=True)
    opset_cols = [oc[1] for oc in opset_cols]
    new_cols = opset_cols[:1]
    bench_cols = ["RT/SKL-N=1", "N=10", "N=100",
                  "N=1000", "N=10000"]
    new_cols.extend(["ERROR-msg", "name", "problem", "scenario", 'optim'])
    new_cols.extend(bench_cols)
    new_cols.extend(opset_cols[1:])
    for c in bench_cols:
        new_cols.append(c + '-min')
        new_cols.append(c + '-max')
    for c in piv.columns:
        if c.startswith("skl_") or c.startswith("onx_"):
            new_cols.append(c)
    new_cols = [_ for _ in new_cols if _ in piv.columns]
    piv = piv[new_cols]

    _save(piv, out_summary)
    _save(df_raw, out_raw)
    return filenames


def validate_runtime(verbose=1, opset_min=-1, opset_max="",
                     check_runtime=True, runtime='python', debug=False,
                     models=None, out_raw="model_onnx_raw.xlsx",
                     out_summary="model_onnx_summary.xlsx",
                     dump_folder=None, dump_all=False, benchmark=False,
                     catch_warnings=True, assume_finite=True,
                     versions=False, skip_models=None,
                     extended_list=True, separate_process=False,
                     time_kwargs=None, n_features=None, fLOG=print,
                     out_graph=None, force_return=False,
                     dtype=None, skip_long_test=False,
                     number=1, repeat=1, time_kwargs_fact='lin',
                     time_limit=4, n_jobs=0):
    """
    Walks through most of :epkg:`scikit-learn` operators
    or model or predictor or transformer, tries to convert
    them into :epkg:`ONNX` and computes the predictions
    with a specific runtime.

    :param verbose: integer from 0 (None) to 2 (full verbose)
    :param opset_min: tries every conversion from this minimum opset,
        -1 to get the current opset
    :param opset_max: tries every conversion up to maximum opset,
        -1 to get the current opset
    :param check_runtime: to check the runtime
        and not only the conversion
    :param runtime: runtime to check, python,
        onnxruntime1 to check :epkg:`onnxruntime`,
        onnxruntime2 to check every *ONNX* node independently
        with onnxruntime, many runtime can be checked at the same time
        if the value is a comma separated list
    :param models: comma separated list of models to test or empty
        string to test them all
    :param skip_models: models to skip
    :param debug: stops whenever an exception is raised,
        only if *separate_process* is False
    :param out_raw: output raw results into this file (excel format)
    :param out_summary: output an aggregated view into this file (excel format)
    :param dump_folder: folder where to dump information (pickle)
        in case of mismatch
    :param dump_all: dumps all models, not only the failing ones
    :param benchmark: run benchmark
    :param catch_warnings: catch warnings
    :param assume_finite: See `config_context
        <https://scikit-learn.org/stable/modules/generated/sklearn.config_context.html>`_,
        If True, validation for finiteness will be skipped, saving time, but leading
        to potential crashes. If False, validation for finiteness will be performed,
        avoiding error.
    :param versions: add columns with versions of used packages,
        :epkg:`numpy`, :epkg:`scikit-learn`, :epkg:`onnx`, :epkg:`onnxruntime`,
        :epkg:`sklearn-onnx`
    :param extended_list: extends the list of :epkg:`scikit-learn` converters
        with converters implemented in this module
    :param separate_process: run every model in a separate process,
        this option must be used to run all model in one row
        even if one of them is crashing
    :param time_kwargs: a dictionary which defines the number of rows and
        the parameter *number* and *repeat* when benchmarking a model,
        the value must follow :epkg:`json` format
    :param n_features: change the default number of features for
        a specific problem, it can also be a comma separated list
    :param force_return: forces the function to return the results,
        used when the results are produces through a separate process
    :param out_graph: image name, to output a graph which summarizes
        a benchmark in case it was run
    :param dtype: '32' or '64' or None for both,
        limits the test to one specific number types
    :param skip_long_test: skips tests for high values of N if
        they seem too long
    :param number: to multiply number values in *time_kwargs*
    :param repeat: to multiply repeat values in *time_kwargs*
    :param time_kwargs_fact: to multiply number and repeat in
        *time_kwargs* depending on the model
        (see :func:`_multiply_time_kwargs <mlprodict.onnxrt.validate.validate_helper._multiply_time_kwargs>`)
    :param time_limit: to stop benchmarking after this limit of time
    :param n_jobs: force the number of jobs to have this value,
        by default, it is equal to the number of CPU
    :param fLOG: logging function

    .. cmdref::
        :title: Validates a runtime against scikit-learn
        :cmd: -m mlprodict validate_runtime --help
        :lid: l-cmd-validate_runtime

        The command walks through all scikit-learn operators,
        tries to convert them, checks the predictions,
        and produces a report.

        Example::

            python -m mlprodict validate_runtime --models LogisticRegression,LinearRegression

        Following example benchmarks models
        :epkg:`sklearn:ensemble:RandomForestRegressor`,
        :epkg:`sklearn:tree:DecisionTreeRegressor`, it compares
        :epkg:`onnxruntime` against :epkg:`scikit-learn` for opset 10.

        ::

            python -m mlprodict validate_runtime -v 1 -o 10 -op 10 -c 1 -r onnxruntime1
                   -m RandomForestRegressor,DecisionTreeRegressor -out bench_onnxruntime.xlsx -b 1

        Parameter ``--time_kwargs`` may be used to reduce or increase
        bencharmak precisions. The following value tells the function
        to run a benchmarks with datasets of 1 or 10 number, to repeat
        a given number of time *number* predictions in one row.
        The total time is divided by :math:`number \\times repeat``.
        Parameter ``--time_kwargs_fact`` may be used to increase these
        number for some specific models. ``'lin'`` multiplies
        by 10 number when the model is linear.

        ::

            -t "{\\"1\\":{\\"number\\":10,\\"repeat\\":10},\\"10\\":{\\"number\\":5,\\"repeat\\":5}}"

        The following example dumps every model in the list:

        ::

            python -m mlprodict validate_runtime --out_raw raw.csv --out_summary sum.csv
                   --models LinearRegression,LogisticRegression,DecisionTreeRegressor,DecisionTreeClassifier
                   -r python,onnxruntime1 -o 10 -op 10 -v 1 -b 1 -dum 1
                   -du model_dump -n 20,100,500 --out_graph benchmark.png --dtype 32

        The command line generates a graph produced by function
        :func:`plot_validate_benchmark
        <mlprodict.onnxrt.validate.validate_graph.plot_validate_benchmark>`.
    """
    if separate_process:
        return _validate_runtime_separate_process(
            verbose=verbose, opset_min=opset_min, opset_max=opset_max,
            check_runtime=check_runtime, runtime=runtime, debug=debug,
            models=models, out_raw=out_raw,
            out_summary=out_summary, dump_all=dump_all,
            dump_folder=dump_folder, benchmark=benchmark,
            catch_warnings=catch_warnings, assume_finite=assume_finite,
            versions=versions, skip_models=skip_models,
            extended_list=extended_list, time_kwargs=time_kwargs,
            n_features=n_features, fLOG=fLOG, force_return=True,
            out_graph=None, dtype=dtype, skip_long_test=skip_long_test,
            time_kwargs_fact=time_kwargs_fact, time_limit=time_limit,
            n_jobs=n_jobs)

    from ..onnxrt.validate import enumerate_validated_operator_opsets  # pylint: disable=E0402

    if not isinstance(models, list):
        models = (None if models in (None, "")
                  else models.strip().split(','))
    if not isinstance(skip_models, list):
        skip_models = ({} if skip_models in (None, "")
                       else skip_models.strip().split(','))
    if verbose <= 1:
        logger = getLogger('skl2onnx')
        logger.disabled = True
    if not dump_folder:
        dump_folder = None
    if dump_folder and not os.path.exists(dump_folder):
        os.mkdir(dump_folder)  # pragma: no cover
    if dump_folder and not os.path.exists(dump_folder):
        raise FileNotFoundError(  # pragma: no cover
            f"Cannot find dump_folder '{dump_folder}'.")

    # handling parameters
    if opset_max == "":
        opset_max = None  # pragma: no cover
    if isinstance(opset_min, str):
        opset_min = int(opset_min)  # pragma: no cover
    if isinstance(opset_max, str):
        opset_max = int(opset_max)
    if isinstance(verbose, str):
        verbose = int(verbose)  # pragma: no cover
    if isinstance(extended_list, str):
        extended_list = extended_list in (
            '1', 'True', 'true')  # pragma: no cover
    if time_kwargs in (None, ''):
        time_kwargs = None
    if isinstance(time_kwargs, str):
        time_kwargs = json.loads(time_kwargs)
        # json only allows string as keys
        time_kwargs = {int(k): v for k, v in time_kwargs.items()}
    if isinstance(n_jobs, str):
        n_jobs = int(n_jobs)
        if n_jobs == 0:
            n_jobs = None
    if time_kwargs is not None and not isinstance(time_kwargs, dict):
        raise ValueError(  # pragma: no cover
            f"time_kwargs must be a dictionary not {type(time_kwargs)}\n{time_kwargs}")
    if not isinstance(n_features, list):
        if n_features in (None, ""):
            n_features = None
        elif ',' in n_features:
            n_features = list(map(int, n_features.split(',')))
        else:
            n_features = int(n_features)
    if not isinstance(runtime, list) and ',' in runtime:
        runtime = runtime.split(',')

    def fct_filter_exp(m, s):
        cl = m.__name__
        if cl in skip_models:
            return False
        pair = f"{cl}[{s}]"
        if pair in skip_models:
            return False
        return True

    if dtype in ('', None):
        fct_filter = fct_filter_exp
    elif dtype == '32':
        def fct_filter_exp2(m, p):
            return fct_filter_exp(m, p) and '64' not in p
        fct_filter = fct_filter_exp2
    elif dtype == '64':  # pragma: no cover
        def fct_filter_exp3(m, p):
            return fct_filter_exp(m, p) and '64' in p
        fct_filter = fct_filter_exp3
    else:
        raise ValueError(  # pragma: no cover
            f"dtype must be empty, 32, 64 not '{dtype}'.")

    # time_kwargs

    if benchmark:
        if time_kwargs is None:
            from ..onnxrt.validate.validate_helper import default_time_kwargs  # pylint: disable=E0402
            time_kwargs = default_time_kwargs()
        for _, v in time_kwargs.items():
            v['number'] *= number
            v['repeat'] *= repeat
        if verbose > 0:
            fLOG(f"time_kwargs={time_kwargs!r}")

    # body

    def build_rows(models_):
        rows = list(enumerate_validated_operator_opsets(
            verbose, models=models_, fLOG=fLOG, runtime=runtime, debug=debug,
            dump_folder=dump_folder, opset_min=opset_min, opset_max=opset_max,
            benchmark=benchmark, assume_finite=assume_finite, versions=versions,
            extended_list=extended_list, time_kwargs=time_kwargs, dump_all=dump_all,
            n_features=n_features, filter_exp=fct_filter,
            skip_long_test=skip_long_test, time_limit=time_limit,
            time_kwargs_fact=time_kwargs_fact, n_jobs=n_jobs))
        return rows

    def catch_build_rows(models_):
        if catch_warnings:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore",
                                      (UserWarning, ConvergenceWarning,
                                       RuntimeWarning, FutureWarning))
                rows = build_rows(models_)
        else:
            rows = build_rows(models_)  # pragma: no cover
        return rows

    rows = catch_build_rows(models)
    res = _finalize(rows, out_raw, out_summary,
                    verbose, models, out_graph, fLOG)
    return res if (force_return or verbose >= 2) else None


def _finalize(rows, out_raw, out_summary, verbose, models, out_graph, fLOG):
    from ..onnxrt.validate import summary_report  # pylint: disable=E0402
    from ..tools.cleaning import clean_error_msg  # pylint: disable=E0402

    # Drops data which cannot be serialized.
    for row in rows:
        keys = []
        for k in row:
            if 'lambda' in k:
                keys.append(k)
        for k in keys:
            del row[k]

    df = DataFrame(rows)

    if out_raw:
        if verbose > 0:
            fLOG(f"Saving raw_data into '{out_raw}'.")
        if os.path.splitext(out_raw)[-1] == ".xlsx":
            df.to_excel(out_raw, index=False)
        else:
            clean_error_msg(df).to_csv(out_raw, index=False)

    if df.shape[0] == 0:
        raise RuntimeError("No result produced by the benchmark.")
    piv = summary_report(df)
    if 'optim' not in piv:
        raise RuntimeError(  # pragma: no cover
            f"Unable to produce a summary. Missing column in \n{piv.columns}")

    if out_summary:
        if verbose > 0:
            fLOG(f"Saving summary into '{out_summary}'.")
        if os.path.splitext(out_summary)[-1] == ".xlsx":
            piv.to_excel(out_summary, index=False)
        else:
            clean_error_msg(piv).to_csv(out_summary, index=False)

    if verbose > 1 and models is not None:
        fLOG(piv.T)
    if out_graph is not None:
        if verbose > 0:
            fLOG(f"Saving graph into '{out_graph}'.")
        from ..plotting.plotting import plot_validate_benchmark
        fig = plot_validate_benchmark(piv)[0]
        fig.savefig(out_graph)

    return rows


def _validate_runtime_dict(kwargs):
    return validate_runtime(**kwargs)


def _validate_runtime_separate_process(**kwargs):
    models = kwargs['models']
    if models in (None, ""):
        from ..onnxrt.validate.validate_helper import sklearn_operators  # pragma: no cover
        models = [_['name']
                  for _ in sklearn_operators(extended=True)]  # pragma: no cover
    elif not isinstance(models, list):
        models = models.strip().split(',')

    skip_models = kwargs['skip_models']
    skip_models = {} if skip_models in (
        None, "") else skip_models.strip().split(',')

    verbose = kwargs['verbose']
    fLOG = kwargs['fLOG']
    all_rows = []
    skls = [m for m in models if m not in skip_models]
    skls.sort()

    if verbose > 0:
        from tqdm import tqdm
        pbar = tqdm(skls)
    else:
        pbar = skls  # pragma: no cover

    for op in pbar:
        if not isinstance(pbar, list):
            pbar.set_description(f"[{op + ' ' * (25 - len(op))}]")

        if kwargs['out_raw']:
            out_raw = os.path.splitext(kwargs['out_raw'])
            out_raw = "".join([out_raw[0], "_", op, out_raw[1]])
        else:
            out_raw = None  # pragma: no cover

        if kwargs['out_summary']:
            out_summary = os.path.splitext(kwargs['out_summary'])
            out_summary = "".join([out_summary[0], "_", op, out_summary[1]])
        else:
            out_summary = None  # pragma: no cover

        new_kwargs = kwargs.copy()
        if 'fLOG' in new_kwargs:
            del new_kwargs['fLOG']
        new_kwargs['out_raw'] = out_raw
        new_kwargs['out_summary'] = out_summary
        new_kwargs['models'] = op
        new_kwargs['verbose'] = 0  # tqdm fails
        new_kwargs['out_graph'] = None

        with Pool(1) as p:
            try:
                result = p.apply_async(_validate_runtime_dict, [new_kwargs])
                lrows = result.get(timeout=150)  # timeout fixed to 150s
                all_rows.extend(lrows)
            except Exception as e:  # pylint: disable=W0703
                all_rows.append({  # pragma: no cover
                    'name': op, 'scenario': 'CRASH',
                    'ERROR-msg': str(e).replace("\n", " -- ")
                })

    return _finalize(all_rows, kwargs['out_raw'], kwargs['out_summary'],
                     verbose, models, kwargs.get('out_graph', None), fLOG)


def latency(model, law='normal', size=1, number=10, repeat=10, max_time=0,
            runtime="onnxruntime", device='cpu', fmt=None,
            profiling=None, profile_output='profiling.csv'):
    """
    Measures the latency of a model (python API).

    :param model: ONNX graph
    :param law: random law used to generate fake inputs
    :param size: batch size, it replaces the first dimension
        of every input if it is left unknown
    :param number: number of calls to measure
    :param repeat: number of times to repeat the experiment
    :param max_time: if it is > 0, it runs as many time during
        that period of time
    :param runtime: available runtime
    :param device: device, `cpu`, `cuda:0` or a list of providers
        `CPUExecutionProvider, CUDAExecutionProvider
    :param fmt: None or `csv`, it then
        returns a string formatted like a csv file
    :param profiling: if True, profile the execution of every
        node, if can be sorted by name or type,
        the value for this parameter should e in `(None, 'name', 'type')`
    :param profile_output: output name for the profiling
        if profiling is specified

    .. cmdref::
        :title: Measures model latency
        :cmd: -m mlprodict latency --help
        :lid: l-cmd-latency

        The command generates random inputs and call many times the
        model on these inputs. It returns the processing time for one
        iteration.

        Example::

            python -m mlprodict latency --model "model.onnx"
    """
    from ..onnxrt.validate.validate_latency import latency as _latency  # pylint: disable=E0402

    if not os.path.exists(model):
        raise FileNotFoundError(  # pragma: no cover
            f"Unable to find model {model!r}.")
    if profiling not in (None, '', 'name', 'type'):
        raise ValueError(  # pragma: no cover
            f"Unexpected value for profiling: {profiling!r}.")
    size = int(size)
    number = int(number)
    repeat = int(repeat)
    if max_time in (None, 0, ""):
        max_time = None
    else:
        max_time = float(max_time)
        if max_time <= 0:
            max_time = None

    if law != "normal":
        raise ValueError(  # pragma: no cover
            f"Only law='normal' is supported, not {law!r}.")

    if profiling in ('name', 'type') and profile_output in (None, ''):
        raise ValueError(  # pragma: no cover
            f'profiling is enabled but profile_output is wrong ({profile_output!r}).')

    res = _latency(
        model, law=law, size=size, number=number, repeat=repeat,
        max_time=max_time, runtime=runtime, device=device,
        profiling=profiling)

    if profiling not in (None, ''):
        res, gr = res
        ext = os.path.splitext(profile_output)[-1]
        gr = gr.reset_index(drop=False)
        if ext == '.csv':
            gr.to_csv(profile_output, index=False)
        elif ext == '.xlsx':  # pragma: no cover
            gr.to_excel(profile_output, index=False)
        else:
            raise ValueError(  # pragma: no cover
                f"Unexpected extension for profile_output={profile_output!r}.")

    if fmt == 'csv':
        st = StringIO()
        df = DataFrame([res])
        df.to_csv(st, index=False)
        return st.getvalue()
    if fmt in (None, ''):
        return res
    raise ValueError(  # pragma: no cover
        f"Unexpected value for fmt: {fmt!r}.")



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class Concat(OpRun):

    atts = {'axis': 0}
    python_inputs = ['*inputs']

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Concat.atts,
                       **options)

    def _preprocess(self, a):
        if len(a.shape) == 0:
            raise RuntimeError(  # pragma: no cover
                f"Concat: one input has an empty shape: {a!r}.")
        if self.axis >= len(a.shape):
            new_shape = a.shape + (1, ) * (self.axis + 1 - len(a.shape))
            return a.reshape(new_shape)
        return a

    def _run(self, *args, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        targs = tuple(self._preprocess(a) for a in args)
        return (numpy.concatenate(targs, self.axis), )

    def to_python(self, inputs):
        return "import numpy", "return numpy.concatenate(inputs, axis=axis)"



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnary
from .op_tfidfvectorizer_ import RuntimeTfIdfVectorizer  # pylint: disable=E0611,E0401


class TfIdfVectorizer(OpRunUnary):

    atts = {'max_gram_length': 1,
            'max_skip_count': 1,
            'min_gram_length': 1,
            'mode': b'TF',
            'ngram_counts': [],
            'ngram_indexes': [],
            'pool_int64s': [],
            'pool_strings': [],
            'weights': []}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnary.__init__(self, onnx_node, desc=desc,
                            expected_attributes=TfIdfVectorizer.atts,
                            **options)
        self.rt_ = RuntimeTfIdfVectorizer()
        if len(self.pool_strings) != 0:
            pool_int64s = list(range(len(self.pool_strings)))
            pool_strings_ = numpy.array(
                [_.decode('utf-8') for _ in self.pool_strings])
            mapping = {}
            for i, w in enumerate(pool_strings_):
                mapping[w] = i
        else:
            mapping = None
            pool_int64s = self.pool_int64s
            pool_strings_ = None

        self.mapping_ = mapping
        self.pool_strings_ = pool_strings_
        self.rt_.init(
            self.max_gram_length, self.max_skip_count, self.min_gram_length,
            self.mode, self.ngram_counts, self.ngram_indexes, pool_int64s,
            self.weights)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.mapping_ is None:
            res = self.rt_.compute(x)
            return (res.reshape((x.shape[0], -1)), )
        else:
            xi = numpy.empty(x.shape, dtype=numpy.int64)
            for i in range(0, x.shape[0]):
                for j in range(0, x.shape[1]):
                    try:
                        xi[i, j] = self.mapping_[x[i, j]]
                    except KeyError:
                        xi[i, j] = -1
            res = self.rt_.compute(xi)
            return (res.reshape((x.shape[0], -1)), )



"""
@file
@brief Custom operators for FFT.
"""
import numpy
from skl2onnx.algebra.onnx_operator import OnnxOperator


class OnnxFFT_1(OnnxOperator):
    """
    Defines a custom operator for FFT.
    """

    since_version = 1
    expected_inputs = [('A', 'T'), ('fft_length', numpy.int64)]
    expected_outputs = [('FFT', 'T')]
    input_range = [1, 2]
    output_range = [1, 1]
    is_deprecated = False
    domain = 'mlprodict'
    operator_name = 'FFT'
    past_version = {}

    def __init__(self, *args, axis=-1,
                 op_version=None, **kwargs):
        """
        :param A: array or OnnxOperatorMixin
        :param fft_length: (optional) array or OnnxOperatorMixin (args)
        :param axis: axis
        :param op_version: opset version
        :param kwargs: additional parameter
        """
        OnnxOperator.__init__(
            self, *args, axis=axis,
            op_version=op_version, **kwargs)


class OnnxFFT2D_1(OnnxOperator):
    """
    Defines a custom operator for FFT2D.
    """

    since_version = 1
    expected_inputs = [('A', 'T'), ('fft_length', numpy.int64)]
    expected_outputs = [('FFT2D', 'T')]
    input_range = [1, 2]
    output_range = [1, 1]
    is_deprecated = False
    domain = 'mlprodict'
    operator_name = 'FFT2D'
    past_version = {}

    def __init__(self, *args, axes=(-2, -1),
                 op_version=None, **kwargs):
        """
        :param A: array or OnnxOperatorMixin
        :param fft_length: (optional) array or OnnxOperatorMixin (args)
        :param axes: axes
        :param op_version: opset version
        :param kwargs: additional parameter
        """
        OnnxOperator.__init__(
            self, *args, axes=axes,
            op_version=op_version, **kwargs)


class OnnxRFFT_1(OnnxOperator):
    """
    Defines a custom operator for FFT.
    """

    since_version = 1
    expected_inputs = [('A', 'T'), ('fft_length', numpy.int64)]
    expected_outputs = [('RFFT', 'T')]
    input_range = [1, 2]
    output_range = [1, 1]
    is_deprecated = False
    domain = 'mlprodict'
    operator_name = 'RFFT'
    past_version = {}

    def __init__(self, *args, axis=-1,
                 op_version=None, **kwargs):
        """
        :param A: array or OnnxOperatorMixin
        :param fft_length: (optional) array or OnnxOperatorMixin (args)
        :param axis: axis
        :param op_version: opset version
        :param kwargs: additional parameter
        """
        OnnxOperator.__init__(
            self, *args, axis=axis,
            op_version=op_version, **kwargs)


OnnxFFT = OnnxFFT_1
OnnxFFT2D = OnnxFFT2D_1
OnnxRFFT = OnnxRFFT_1



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from ._op_helper import proto2dtype, dtype_name


class EyeLike(OpRun):

    atts = {'k': 0, 'dtype': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=EyeLike.atts,
                       **options)
        self.dtype_ = proto2dtype(self.dtype)

    def _run(self, data, *args, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        shape = data.shape
        if len(shape) == 1:
            sh = (shape[0], shape[0])
        elif len(shape) == 2:
            sh = shape
        else:
            raise RuntimeError(  # pragma: no cover
                f"EyeLike only accept 1D or 2D tensors not {shape!r}.")
        return (numpy.eye(*sh, k=self.k, dtype=self.dtype_), )

    def to_python(self, inputs):
        return (
            "import numpy",
            "return numpy.eye(*(%s.shape), k=%d, dtype=numpy.%s)" % (
                inputs[0], self.k, dtype_name(self.dtype_)))



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class Range(OpRun):

    atts = {}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Range.atts,
                       **options)

    def _run(self, starts, ends, steps, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return (numpy.arange(starts, ends, steps).astype(starts.dtype), )
